/*
 * Copyright (c) 1995 Boris V. Tobotras
 * boris@xtalk.msk.su AKA 2:5020/20.7 AKA 2:5020/188.21
 *
 * Created: Fri Feb  3 14:28:16 1995
 *
 * Txy flag (FSC-0062) processing functions for ifcico
 *
 * Consider code below as freeware, please.
 *
 */
# include	<ctype.h>
# include	<time.h>
# include	"lutil.h"
# include	"Txy.h"
# include	<string.h>
# include	<stdlib.h>
#ifdef IFCS
#include "ifcs.h"
void ipcloginf(int, char*,...);
#endif

int checkWorkTime(node *, char *, int, int, int);
int checkTxy(char *, int, int);
int checkTimeRanges(char *, int, int, int);

int checkWorkTime( node *nlent, char *timeRanges, int Day, int Hour, int Min )
{
	int i;
	
	if(timeRanges) return checkTimeRanges(timeRanges, Day, Hour, Min);

	if ( nlent->flags & CM ) {
		debug( 10, "Calling CM node" );
		return 1;
	}

	if ( (nlent->flags & MH01) && checkTxy("BC", Hour, Min )) {
		debug( 10, "Calling #01-flag node" );
		return 1;
	}

	if ( (nlent->flags & MH02) && checkTxy("cd", Hour, Min )) {
		debug( 10, "Calling #02-flag node" );
		return 1;
	}

	if ( (nlent->flags & MH08) && checkTxy("IJ", Hour, Min )) {
		debug( 10, "Calling #08-flag node" );
		return 1;
	}

	if ( (nlent->flags & MH09) && checkTxy("JK", Hour, Min )) {
		debug( 10, "Calling #09-flag node" );
		return 1;
	}

	if ( (nlent->flags & MH18) && checkTxy("ST", Hour, Min )) {
		debug( 10, "Calling #18-flag node" );
		return 1;
	}

	if ( (nlent->flags & MH20) && checkTxy("UV", Hour, Min )) {
		debug( 10, "Calling #20-flag node" );
		return 1;
	}
	
	for ( i = 0; i < MAXUFLAGS; ++i ) {
		char *flag = nlent->uflags[ i ];
		if ( flag ) {
			if ( flag[ 0 ] == 'T' && flag[ 3 ] == '\0' ) {
				debug( 10, "Found Txy flag: %s", flag++ );
				return checkTxy( flag, Hour, Min );
			} else if ( flag[ 0 ] == 'U' && flag[ 1 ] == 'T' &&
				      flag[ 4 ] == '\0' ) {
				flag++;
				debug( 10, "Found Txy flag: %s", flag++ );
				return checkTxy( flag, Hour, Min );
			}
		}
		else break;
	}
	debug( 10, "No Txy flag" );
	/* No Txy flag... defaulting to ZMH */
	/* points don't have to honour ZMH */
	if (nlent->addr.point != 0) {
		if (nlent->addr.zone==1) {
			debug(10,"Using Z1MH");
			return checkTxy( "JK", Hour, Min );
		} else if (nlent->addr.zone==2) {
			debug(10,"Using Z2MH");
			return checkTxy( "cd", Hour, Min );
		} else if (nlent->addr.zone==3) {
			debug(10,"Using Z3MH");
			return checkTxy( "ST", Hour, Min );
		} else if (nlent->addr.zone==4) {
			debug(10,"Using Z4MH");
			return checkTxy( "IJ", Hour, Min );
		} else if (nlent->addr.zone==5) {
			debug(10,"Using Z5MH");
			return checkTxy( "BC", Hour, Min );
		} else if (nlent->addr.zone==6) {
			debug(10,"Using Z6MH");
			return checkTxy( "UV", Hour, Min );
		} else {
			debug( 10, "node to call is not on fidonet");
		}
	}
return 0;
}

int checkTxy(char *xy, int forcedHour, int forcedMin)
{
	time_t GMTime = time( NULL );
	time_t LocalTime;
	struct tm *Now;
	struct tm *Local = localtime( &GMTime );
	int NowMin, WorkTimeStartMin, WorkTimeEndMin;
	
	if(forcedHour != -1 && forcedMin != -1)
	    {
	    Local->tm_hour = forcedHour;
	    Local->tm_min = forcedMin;
	    LocalTime = mktime(Local);
	    Now = gmtime(&LocalTime);
	    }
	else Now = gmtime( &GMTime );
	
	WorkTimeStartMin = ( toupper( *xy ) - 'A' ) * 60;
	if ( islower( *xy++ ) )	WorkTimeStartMin += 30;
	WorkTimeEndMin = ( toupper( *xy ) - 'A' ) * 60;
	if ( islower( *xy ) ) WorkTimeEndMin += 30;

	NowMin = Now->tm_hour * 60 + Now->tm_min;
	
	debug( 10, "Working time: %d:%d-%d:%d, now %d:%d GMT", 
		WorkTimeStartMin / 60, WorkTimeStartMin % 60,
		WorkTimeEndMin / 60, WorkTimeEndMin % 60,
		NowMin / 60, NowMin % 60 ); 
	
	if ( WorkTimeEndMin > WorkTimeStartMin )
		if ( NowMin >= WorkTimeStartMin && NowMin < WorkTimeEndMin )
			return 1;
	    else
			return 0;
	else
		if ( NowMin >= WorkTimeStartMin || NowMin < WorkTimeEndMin )
			return 1;
		else
			return 0;
}


int checkTimeRanges(char *timeRanges, int forcedDay, int forcedHour,
							int forcedMin)
{
int firstDay, secondDay, firstHour, firstMinute, secondHour, secondMinute,
    firstMark, secondMark, currentMark, Day, Hour, Min;
char *p, *s, *f, *rs, *r;
time_t Time;
struct tm *localTime;

if(!strlen(timeRanges))
    {
    loginf("Empty timerange string!");
    return 0;
    }
    
if(!(rs = strdup(timeRanges))) return 0;	// 

debug(10, "work timeranges: \"%s\"", rs);

for(r = strtok(rs, ","); r; r = strtok(NULL, ","))
    {
    if(!strcasecmp(r, "CM")) { free(rs); return 1;}
    if(!strlen(r)) continue;

    firstDay = -1;
    secondDay = -1;
    firstHour = -1;
    firstMinute = -1;
    secondHour = -1;
    secondMinute = -1;

    debug(10,"timerange: \"%s\"", r);    

    if((p = strchr(r, '-')))	 //     ? (1)
	{			 // . (1)
        if((s = strchr(r, '.'))) //    () ? (2)
	    {			 // . (2)
	    firstDay = atoi(s - 1);
	    firstHour = atoi(s + 1);
	    if((f = strchr(s + 1, ':')) && f < p) firstMinute = atoi(f + 1);
	
	    if((s = strchr(p + 1, '.')))
		{
		secondDay = atoi(s - 1);
		secondHour = atoi(s + 1);
		if((f = strchr(s + 1, ':'))) secondMinute = atoi(f + 1);
		}
	    else
		{
		logerr("Error parsing timerange: \"%s\"", r);
		free(rs);
		return 0;	//    .
		}
	    }
	else
	    {			// . (2)
	    firstHour = atoi(r);
	    if((f = strchr(r, ':')) && f < p) firstMinute = atoi(f + 1);
	    secondHour = atoi(p + 1);
	    if((f = strchr(p + 1, ':'))) secondMinute = atoi(f + 1);
	    }
	}
    else
	{			    // . (1)
	if((s = strchr(r, '.'))) //    () ? (3)
	    {			    // . (3)
	    firstDay = atoi(s - 1);
	    firstHour = atoi(s + 1);
	    if((p = strchr(s + 1,':'))) firstMinute = atoi(p + 1);
	    }
	else
	    {			// . (3)
	    firstHour = atoi(r);
	    if((p = strchr(r,':'))) firstMinute = atoi(p + 1);
	    }
	}
    
    if( firstDay < -1 || firstDay > 7 || firstDay == 0 || firstHour < -1 ||
	firstHour > 23 || firstMinute< -1 || firstMinute > 59 || secondDay < -1
	|| secondDay > 7  || secondDay == 0 || secondHour < -1 ||
	secondHour > 24 || secondMinute < -1 || secondMinute > 59 ||
	firstHour == -1 || (secondDay != -1 && secondHour == -1))
	{
	logerr("Error parsing timerange: \"%s\"", r);
	free(rs);
	return 0;	//   .
	}

    debug(10,"parsed timerange: %i.%i:%i - %i.%i:%i", firstDay, firstHour,
	    firstMinute, secondDay, secondHour, secondMinute);
   
    Time = time(NULL);			//  
    localTime = localtime(&Time);	//    

    if(forcedDay == -1)	Day = localTime->tm_wday;
    else Day = forcedDay;
    if(!Day) Day = 7;
    if(forcedHour == -1) Hour = localTime->tm_hour;
    else Hour = forcedHour; 
    if(forcedMin == -1)	Min = localTime->tm_min;
    else Min = forcedMin;

    firstMark = firstHour * 60;	//     
    if(firstMinute != -1) firstMark += firstMinute;

    secondMark = secondHour * 60;	//      
    if(secondMinute != -1) secondMark += secondMinute;

    currentMark = Hour * 60 + Min;	//    
    
    debug(10, "firstmark: %i, secondmark: %i, currentmark: %i", firstMark,
		secondMark, currentMark);
	
    if(secondDay != -1)	//     
        {
	if(firstDay < secondDay)
	    {
	    if(Day >= firstDay && Day <= secondDay)
		{
		if(firstMark < secondMark)
		    {
		    if(currentMark >= firstMark && currentMark < secondMark)
			{ free(rs); return 1; }
		    }
		else
		    {
		    if(currentMark >= firstMark || currentMark < secondMark)
			{ free(rs); return 1; }
		    }
		}
	    }
	else 
	    {
	    if(firstDay == secondDay)
		{
		if(Day == firstDay)
		    {
		    if(firstMark < secondMark)
			{
			if(currentMark >= firstMark && currentMark < secondMark)
			    { free(rs); return 1;}
			}
		    else
			{
			if(currentMark >= firstMark || currentMark < secondMark)
			    { free(rs); return 1;}
			}
		    }
		}
	    else
		{
		if(Day >= firstDay || Day <= secondDay)
		    {
		    if(firstMark < secondMark)
			{
			if(currentMark >= firstMark && currentMark < secondMark)
			    { free(rs); return 1;}
			}
		    else
			{
			if(currentMark >= firstMark || currentMark < secondMark)
			    { free(rs); return 1;}
			}
		    }
		}
	    }
	}
    else if(secondHour != -1)	//   
	    {
	    if(firstMark < secondMark)
		{
	        if(currentMark >= firstMark && currentMark < secondMark)
		    { free(rs); return 1;}
		}
	    else
		{
		if(currentMark >= firstMark || currentMark < secondMark)
		    { free(rs); return 1;}
		}
	    }
	else if(firstDay != -1)		//    
		{
		if(firstMinute != -1)
		    {
		    if(Day == firstDay && Hour == firstHour && Min == firstMinute)
			{ free(rs); return 1;}
		    }
		else
		    {
		    if(Day == firstDay && Hour == firstHour)
			{ free(rs); return 1;}
		    }
		}
	    else		//   
		{
		if(firstMinute != -1)
		    {
		    if(Hour == firstHour && Min == firstMinute)
			{ free(rs); return 1;}
		    }
		else
		    {
		    if(Hour == firstHour)
			{ free(rs); return 1;}
		    }
		}
    }

free(rs);
debug(10, "working range not found");
return 0;
}
