#ifndef FSM_H
#define FSM_H

/*
   Finite State Machine.  Replace oldish "statetbl.h", use '#'
   preprocessor feature, detach state machine from function definition.
   Although, multi-machine (with more than one state variable)
   not implemented, so bidirectional protos will not use this.
*/

#define SM_STATES(name) \
	int sm_success=0;\
	char *sm_name=name;\
	enum {

#define SM_EDECL \
	} sm_state;
			
#define SM_START(x) \
	sm_state=x;\
	debug(14,"statemachine %s start %s (%d)",sm_name,#x,x);\
	while (!sm_success) switch (sm_state)\
	{\
	default: logerr("statemachine %s error: state=%d",sm_name,sm_state);\
		sm_success=-1;

#define SM_STATE(x) \
	break;\
	case x: debug(15,"statemachine %s entering %s (%d)",\
		sm_name,#x,x);

#define SM_END \
	}\
	debug(14,"statemachine %s exit %s (%d)",\
		sm_name,(sm_success == -1)?"error":"success",sm_success);

#define SM_RETURN \
	return (sm_success != 1);

#define SM_PROCEED(x) \
	sm_state=x; break;

#define SM_SUCCESS \
	sm_success=1; break;

#define SM_ERROR \
	sm_success=-1; break;

	
/* Extension for bidirection protos */	
#define BSM_NAME(name, scase) \
	int		case_state=scase;\
	char	*bsm_name=name;

#define BSM_STATES \
	int		rx_success=0,\
			tx_success=0;\
	enum	{
			
#define BSM_EDECL \
			, _dummy_state\
	} tx_state, rx_state;
	
#define BSM_START(x, y) \
	tx_state=x; rx_state=y;\
	debug(14,"statemachine %s start tx=%s (%d), rx=%s (%d)",bsm_name,#x,x, #y, y);\
	while (!(tx_success && rx_success)) { \
	switch (case_state ? tx_state : rx_state)\
	{\
	case _dummy_state: \
		debug(14, "statemachine %s dummy %sstate", bsm_name, case_state ? "tx" : "rx");\
			break;\
	default: logerr("statemachine %s error: %cx state=%d", bsm_name, \
		case_state ? 't' : 'r', case_state ? tx_state : rx_state);\
		case_state ? tx_success : rx_success = -1;\
		goto bsm_end;
			 
#define TX_STATE(x) \
	break;\
	case x: debug(15,"statemachine %s entering tx state %s (%d)",\
		bsm_name,#x,x);

#define TRX_STATE(x, y) \
	break;\
	case x: if (case_state)  debug(15,"statemachine %s entering tx state %s (%d)",\
		bsm_name,#x,x);\
	case y: if (!case_state) debug(15,"statemachine %s entering rx state %s (%d)",\
		bsm_name,#y,y);
			

#define RX_STATE(x) \
	break;\
	case x: debug(15,"statemachine %s entering rx state %s (%d)",\
		bsm_name,#x,x);

			
#define BSM_END };\
	case_state = !case_state;\
	}\
bsm_end:\
	debug(14,"statemachine %s exit %s (%d %d)",\
		bsm_name,(tx_success == 1 && rx_success == 1 ) ? "success":"error",\
		tx_success, rx_success);

#define BSM_RETURN \
	return !(tx_success == 1 && rx_success == 1);

#define TX_PROCEED(x) \
	{ tx_state=x; break; }

#define RX_PROCEED(x) \
	{ rx_state=x; break; }

#define RX_SUCCESS(x) \
	{ rx_success=1; rx_state = x; break; }

#define TX_SUCCESS(x) \
	{ tx_success=1; tx_state = x; break; }

#define TX_ERROR(x) \
	{ tx_success = -1; tx_state = x; break; }

#define RX_ERROR(x) \
	{ rx_success = -1; rx_state = x; break; }

#define BSM_ERROR \
	{ tx_success = -1; rx_success = -1; break; }

#endif
