#include <sys/types.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <signal.h>
#ifdef HAS_SYSLOG
#include <syslog.h>
#endif
#include "lgetopt.h"
#include "lutil.h"
#include "xutil.h"
#include "ftn.h"
#include "nodelist.h"
#include "trap.h"
#include "config.h"
#include "version.h"
#ifdef SHOW_SPEED
#include "ttyio.h"
#endif

#if defined(SHORT_PID_T)
#define pid_t short
#elif defined(INT_PID_T)
#define pid_t int
#endif

#ifdef IFCS
#include <varargs.h>
#include <errno.h>
#include "ifcs.h"
#endif

unsigned int call_line, call_lines;

int master=0;
int forcedcalls=0;
char *forcedphone=NULL;
char *forcedline=NULL;
char *inetaddr=NULL;
#if defined(HAS_TCP) || defined(HAS_TERM)
int telnet_mode=0;
#endif
#ifdef SHOW_SPEED
extern char *inconn;
#endif
#ifdef NOISEDEBUG
int junklevel=0;
#endif

int call(faddr *);
fa_list *callall(void);
int answer(char *);
void mkdirs(char*);
void setargspace(char**,char**);
void setproctitle(char*);

#ifdef IFCS
void ShmExit (void);
void ShmReset (void);

shmHead_ *shmHead=NULL;
shmData_ *shmData=NULL;
int shmDataId, msgId=-1;
key_t shmDataKey;
unsigned char shmLine;
struct shmid_ds shmStat;
struct msqid_ds msgStat;

unsigned long SentSize, SentTime, SentCPS, RecvSize, RecvTime, RecvCPS;

void ipcloginf(va_alist)
//va_dcl
{
va_list	args;
char	*fmt, Buff[MAX_LOG+1];
struct tm *TM;
time_t TT;
struct msg_
 {
 long mtype;
 char mtext[MAX_LOG+1];
 } msg;
int color;

if(msgId<0) return;

time(&TT);
TM=localtime(&TT);

va_start(args);
color=va_arg(args, int);
fmt=va_arg(args, char*);

vsnprintf(Buff, MAX_LOG-6, *fmt == '$' ? fmt+1 : fmt, args);
sprintf(msg.mtext, "%02u:%02u ", TM->tm_hour, TM->tm_min);
strcat(msg.mtext, Buff);

msg.mtype=color;

if(msgsnd(msgId, &msg, strlen(msg.mtext)+1, IPC_NOWAIT)<0)
 logerr("Cant send msg to Log queue, msgId %i (%s), mtype %i, mtext %s",
 msg.mtype, msg.mtext, msgId, strerror(errno));

va_end(args);
}
#endif

void usage(void)
{
char validopts[256];
	validopts[0] = '\0';

#if defined(HAS_TCP) || defined(HAS_TERM)
	strcat( validopts, "-a<inetaddr> " );
#endif

#ifdef NOICEDEBUG
	strcat( validopts, "-j<num> " );
#endif
#ifdef SHOW_SPEED
	strcat( validopts, "-S<string> " );
#endif
	strcat( validopts, "-n<phone> -l<ttydevice> -r<role> <node> ..." );
	confusage( validopts );

	fprintf(stderr,"-r 0|1		1 - master, 0 - slave	[0]\n");
	fprintf(stderr,"-n<phone>	forced phone number\n");
	fprintf(stderr,"-l<ttydevice>	forced tty device\n");
#ifdef NOICEDEBUG
	fprintf(stderr,"-j<num>         damage every <num> byte [%d]\n",
								junklevel);
#endif
#if defined(HAS_TCP) || defined(HAS_TERM)
	fprintf(stderr,"-a<inetaddr>	use TCP/IP instead of modem\n");
	fprintf(stderr,"-t<mode>	1 - telnet, 0 - transparent	[0]\n");
#endif
#ifdef SHOW_SPEED
	fprintf(stderr,"-S<connect_string>   show this string into log file\n");
#endif
	fprintf(stderr,"  <node>	should be in domain form, e.g. f11.n22.z3\n");
	fprintf(stderr,"		(this implies master mode)\n");
	fprintf(stderr,"\n or: %s tsync|yoohoo|**EMSI_INQC816\n",myname);
	fprintf(stderr,"		(this implies slave mode)\n");
}

int main(argc,argv,envp)
int argc;
char *argv[];
char *envp[];
{
	int c,uid;
	fa_list *callist=NULL,**tmpl;
	faddr *tmp;
	int rc,maxrc,callno=0,succno=0;
	char *answermode=NULL,*p;
	pid_t child,waitret;
	int status;
	char validopts[256];

#ifdef IFCS
	char *shmKeyFile=IFCS;
	key_t shmHeadKey, shmI;
	int shmHeadId;
	
	time_t timS;
	node *nlent;
#endif

	setargspace(argv,envp);

#if defined(HAS_SYSLOG) && defined(CICOLOG)
	logfacility=CICOLOG;
#endif

	setmyname(argv[0]);
	catch(myname);
	strcpy( validopts, "x:r:n:l:I:h" );
#ifdef NOISEDEBUG
	strcat( validopts, "j:" );
#endif
#if defined(HAS_TCP) || defined(HAS_TERM)
	strcat( validopts, "a:" );
#endif
#ifdef SHOW_SPEED
	strcat( validopts, "S:" );
	inconn = getenv("CONNECT");
#endif
	while ((c=getopt(argc,argv,validopts)) != -1)
	if (confopt(c,optarg)) switch (c)
	{
#ifdef NOISEDEBUG
	case 'j':	junklevel=atoi(optarg); break;
#endif
	case 'r':	master=atoi(optarg);
			if ((master != 0) && (master != 1))
			{
				usage();
				exit(1);
			}
			break;
	case 'l':	forcedline=xstrcpy(optarg); break;
#if defined(HAS_TCP) || defined(HAS_TERM)
	case 'a':	inetaddr=xstrcpy(optarg); break;
	case 't':	telnet_mode=atoi(optarg); break;
#endif
#ifdef SHOW_SPEED
	case 'S':	inconn=optarg; break;
#endif
	case 'n':	forcedphone=xstrcpy(optarg); break;
	default:	usage(); exit(1);
	}

	if (readconfig())
	{
		fprintf(stderr,"Error getting configuration, aborting\n");
		exit(1);
	}

#ifdef IFCS
shmHead = NULL;

if(ipcfile) shmKeyFile = ipcfile;

debug(6,"shm keyfile \"%s\"", shmKeyFile);

if ((shmHeadKey = ftok(shmKeyFile, 0)) < 0)
    logerr ("Error found keyfile (%s) for shared memory operation (header)", shmKeyFile);
else 
    {
    if((shmHeadId = shmget(shmHeadKey, sizeof(shmHead_), 0)) < 0 )
	{
	logerr("Unable to get shared memory segment, try to creat (header)");
	if((shmHeadId = shmget(shmHeadKey, sizeof(shmHead_), IPC_CREAT | 0664)) < 0)
	    {
	    logerr("Unable to creat shared memory segment (header)");
	    }
	else
	    {
	    if((int) (shmHead = shmat(shmHeadId, 0, 0)) <0)
		{
		logerr("Unable to attach shared memory segment (header)");
		shmHead = NULL;
		}
	    else
		{
		for(shmI=1; shmI <= UCHAR_MAX; shmI++)
		    {
		    shmHead->Key[shmI] = -1;
		    }
		}
	    }
	}
    else
	{
	if((int) (shmHead = shmat(shmHeadId, 0, 0)) <0)
	    {
	    logerr("Unable to attach shared memory segment (header)");
	    shmHead = NULL;
	    }
	}
    }

shmData = NULL;

if (shmHead)
    {
nextsocket:
    debug(6,"selecting ifcs socket");
    for(shmI = 1; shmI < UCHAR_MAX; shmI++)
	if(shmHead->Key[shmI] == -1) break;
	else
	    if(shmHead->Key[shmI] == -2) debug(6,"ifcs socket %u damaged", shmI);
	    else
	        {
	        debug(6,"ifcs socket %u busy, checking him", shmI);
		if((shmDataId = shmget(shmHead->Key[shmI], sizeof(shmData_), 0664)) < 0)
	    	    {
		    debug(6,"unable to get id for socket %u", shmI);
		    shmHead->Key[shmI]= -2;
		    debug(6, "socket %u marked as damaged", shmI);
		    goto nextsocket;
	    	    }
		    else
			if((shmData = shmat(shmDataId, 0, 0)) < 0 )
			    {
			    debug(6, "unable to attach to socket %u", shmI);
			    shmData = NULL;
			    shmHead->Key[shmI]= -2;
			    debug(6, "socket %u marked as damaged", shmI);
			    goto nextsocket;
			    }
			else
			    {
			    if(kill(shmData->Pid,0))
				{
				if(shmdt ((char *) shmData) < 0)
			    	    debug(6, "error detach from socket %u", shmI);
				shmData = NULL;
				shmHead->Key[shmI]= -2;
				debug(6, "socket %u marked as damaged", shmI);
				goto nextsocket;
				}
			    else 
			        {
				debug(6, "socket %u is ok, detaching", shmI);
			        if(shmdt ((char *) shmData) < 0)
				    {
			    	    debug(6, "error detach from socket %u", shmI);
				    shmData = NULL;
				    shmHead->Key[shmI]= -2;
				    debug(6, "socket %u marked as damaged", shmI);
				    goto nextsocket;
				    }
			        }			
			    }
		}
	
    if(shmI < UCHAR_MAX)
	{
	debug(6,"selected ifcs socket %u",shmI);
	
	if((shmDataKey = ftok(shmKeyFile, shmI)) <0 )
	    {
	    logerr("ifcs: Error found keyfile (%s) for shared memory operation (data)", shmKeyFile);
	    }
	else
	    {
	    if((shmDataId = shmget(shmDataKey, sizeof(shmData_), IPC_CREAT | 0664)) < 0)
	        {
	        logerr("ifcs: Unable to creat shared memory segment (data)");
	        }
	    else
		{
		if((shmData = shmat(shmDataId, 0, 0)) < 0 )
		    {
		    logerr("ifcs: Unable to attach shared memory segment (data)");
		    shmData = NULL;
		    }
		else
		    {
		    shmData->Pid = getpid();
		    shmHead->Key[shmI] = shmDataKey;
		    shmLine = shmI;

		    if((msgId=msgget(shmDataKey, IPC_CREAT | 0644))<0)
			logerr("ifcs: Cant open or creat Log msg queue (%s)", strerror(errno));

		    debug(6,"attached to ifcs socket %u, pid %u, shmkey %u, shmid %u", shmI,
		    getpid(), shmDataKey, shmDataId);
		    }
		}
	    }
	if(!shmData)
	    { 
	    shmHead->Key[shmI] = -2; 
	    debug(6, "socket %u marked as damaged", shmI);
	    goto nextsocket; 
	    }
	}
    else
	{
	logerr("ifcs: more than %u lines detected", UCHAR_MAX-1);
	}
    }

timS = time (&timS);

if(shmData)
    {
    ShmReset();
    
    atexit(ShmExit);
    
    if (master)
	shmData->SesMaster = 1;
    else
	shmData->SesMaster = 0;
	
    if(inconn)
	{
	if(snprintf(shmData->ConnectStr, MAX_CONNECT, inconn)<0)
	    {
	    shmData->ConnectStr[MAX_CONNECT-1]='>';
	    shmData->ConnectStr[MAX_CONNECT]='\0';
	    }
	}
    if(snprintf(shmData->Whoami, MAX_FTN_ADDR, "%s@%s", ascfnode(whoami->addr, 0x0f),
							whoami->addr->domain)<0)
	{
	shmData->Whoami[MAX_FTN_ADDR-1]='>';
	shmData->Whoami[MAX_FTN_ADDR]='\0';
	}
    if(snprintf(shmData->Ver, MAX_IFCICO_VER, version)<0)
	{
	shmData->Ver[MAX_IFCICO_VER-1]='>';
	shmData->Ver[MAX_IFCICO_VER]='\0';
	}
    }
else logerr("ifcs: ifcs support disabled");
#endif

	maxpsize=0L; /* we want classic packet names here */

	tmpl=&callist;

	while (argv[optind])
	{
		for (p=argv[optind];(*p) && (*p == '*');p++);
		if ((strcasecmp(argv[optind],"tsync") == 0) ||
		    (strcasecmp(argv[optind],"yoohoo") == 0) ||
		    (strncasecmp(p,"EMSI_",5) == 0))
		{
			master=0;
			answermode=argv[optind];
			debug(10,"inbound \"%s\" mode",S(answermode));
		}
		else
		{
			debug(8,"callist entry \"%s\"",argv[optind]);
			if ((tmp=parsefaddr(argv[optind])))
			{
				*tmpl=(fa_list *)xmalloc(sizeof(fa_list));
				(*tmpl)->next=NULL;
				(*tmpl)->addr=tmp;
				tmpl=&((*tmpl)->next);
			}
			else logerr("unrecognizable address \"%s\"",argv[optind]);
		}
		optind++;
	}

	if (callist)
	{
		master=1;
		forcedcalls=1;
#ifdef IFCS
		if(shmData) shmData->SesMaster = 1;
#endif
	}

	/*
	   The following witchkraft about uid-s is necessary to make
	   access() work right.  Unforunately, access() checks the real
	   uid, if ifcico is invoked with supervisor real uid (as when
	   called by uugetty) it returns X_OK for the magic files that
	   even do not have `x' bit set.  Therefore, `reference' magic
	   file requests are taken for `execute' requests (and the
	   actual execution natually fails).  Here we set real uid equal
	   to effective.  If real uid is not zero, all these fails, but
	   in this case it is not necessary anyway.
	*/

	uid=geteuid();
	seteuid(0);
	setuid(uid);
	seteuid(uid);
	debug(2,"uid=%d, euid=%d",getuid(),geteuid());

	umask(066); /* packets may contain confidential information */

#if(defined(BSD_SIGNALS) || defined(HAS_BSD_SIGNALS))
	siginterrupt(SIGALRM,1);
	siginterrupt(SIGPIPE,1);
#endif

	p=xstrcpy(inbound);
	p=xstrcat(p,"/tmp/");
	mkdirs(p);
	free(p);
	p=xstrcpy(listinbound);
	p=xstrcat(p,"/tmp/");
	mkdirs(p);
	free(p);
	p=xstrcpy(protinbound);
	p=xstrcat(p,"/tmp/");
	mkdirs(p);
	free(p);

	maxrc=0;
	if (master)
	{

		if ((child=fork()))
		{
			if (child == -1)
			{
				logerr("$fork() error");
				exit(1);
			}
#ifdef IFCS
	        if(shmData) shmData->Pid = child;
#endif
			while (((waitret=wait(&status)) != -1) &&
			       (waitret != child))
			{
				logerr("wait return %d, status %d,%d",
					waitret,status>>8,status&0xff);
			}
			if (status&0xff) kill(getpid(),status&0xff);
			else exit(status>>8);
		}

		if (callist == NULL) callist=callall();
		for (tmpl=&callist;*tmpl;tmpl=&((*tmpl)->next))
		{
			callno++;
#ifdef IFCS
			if(shmData)
			    {
			ShmReset();
			if(inetaddr) 
			    {
			    if(snprintf(shmData->OtherEndPhone, MAX_PHONE, inetaddr)<0)
				{
				shmData->OtherEndPhone[MAX_PHONE-1]='>';
				shmData->OtherEndPhone[MAX_PHONE]='\0';
				}
			    }
			    }
#endif
			rc=call((*tmpl)->addr);
			loginf("call to %s %s (rc=%d)",
				ascfnode((*tmpl)->addr,0x1f),
				rc?"failed":"successful",rc);
#ifdef IFCS
			nlent=getnlent((*tmpl)->addr);
			ipcloginf(5, "call to %s %s",
				ascfnode((*tmpl)->addr,0x1f),
				rc?"failed":"successful");
			if(shmData) ShmReset();
#endif
			if (rc > maxrc) maxrc=rc;
			if (rc == 0) succno++;
		}
	}
	else /* slave */
	{
#ifdef IFCS
		if(shmData) ShmReset();
#endif
		setproctitle("-ifcico answering");
		maxrc=answer(answermode);
		callno=1;
		succno=(maxrc == 0);
	}

	if (callno)
		loginf("%d of %d calls, maxrc=%d",succno,callno,maxrc);

	return maxrc;
}
#ifdef IFCS
void ShmExit (void)
{
if(shmctl(shmDataId, IPC_STAT, &shmStat) < 0)
 logerr("Error stat shm segment (%s)", strerror(errno));
if(shmctl(shmDataId, IPC_RMID, &shmStat) < 0)
 logerr("Error release shm segment (%s)", strerror(errno));
if(shmStat.shm_nattch == 1) shmHead->Key[shmLine] = -1;    
if(shmdt ((char *) shmData) < 0)
 logerr("Error detach from shm segment (%s)", strerror(errno));

if(!(msgId<0)) if(msgctl(msgId, IPC_RMID, NULL)<0);
}

void ShmReset (void)
{
shmData->tty[0]='\0';
shmData->ttyspeed=0;
 
shmData->ConnectStr[0] = '\0';
shmData->OtherEndPhone[0] = '\0';
shmData->OtherEndLine=0;
shmData->OtherEndLines=0;
  
shmData->EMSI_Addr[0] = '\0';
shmData->EMSI_System[0] = '\0';
shmData->EMSI_SySop[0] = '\0';
shmData->EMSI_Using[0] = '\0';
shmData->EMSI_Location[0] = '\0';
shmData->EMSI_Phone[0] = '\0';
shmData->EMSI_Baud[0] = '\0';
shmData->EMSI_Flags[0] = '\0';
shmData->EMSI_AKA[0] = '\0';
shmData->EMSI_Time = (time_t) 0;

shmData->SesStart = (time_t) 0;
shmData->SesProtected = 0;

shmData->send.TrStart = (time_t) 0;
shmData->send.TrTime = (time_t) 0;
shmData->send.TrSize = 0L;
shmData->send.TrFiles = 0L;
shmData->send.TrSkip = 0L;
shmData->send.Start = (time_t) 0;
shmData->send.FileName[0] = '\0';
shmData->send.FileSize = 0L;
shmData->send.SkipSize = 0L;
shmData->send.Pos = 0L;
shmData->send.BlockSize = 0L;

shmData->recv.TrStart = (time_t) 0;
shmData->send.TrTime = (time_t) 0;
shmData->recv.TrSize = 0L;
shmData->recv.TrFiles = 0L;
shmData->recv.TrSkip = 0L;
shmData->recv.Start = (time_t) 0;
shmData->recv.FileName[0] = '\0';
shmData->recv.FileSize = 0L;
shmData->recv.SkipSize = 0L;
shmData->recv.Pos = 0L;
shmData->recv.BlockSize = 0L;
}
#endif
