#ifdef IFCS
#include <stdio.h>
#include <stdlib.h>
#include <ncurses.h>
#include <signal.h>
#include <time.h>
#include <string.h>
#include <limits.h>
#include <errno.h>
#include <sys/errno.h>
#include <sys/types.h>
#include "ifcs.h"
#include <termios.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <sys/time.h>
#include <linux/serial.h>

#include "lutil.h"
#include "ftn.h"
#include "version.h"
#include "config.h"

#define NORMAL 0
#define SHM_ERROR 1
#define NCURSES_ERROR 2
#define CONFIG_ERROR 3

typedef struct _Win
    {
    int x, y, cols, rows;
    WINDOW *w;
    }
Win;

void PrintLog (Win);
void PrintTransfer (Win);
void PrintEMSIData (Win);
void AtExit(void);
void usage(void);

shmData_ *shmData=NULL;
shmHead_ *shmHead=NULL;
unsigned char shmLine=0, shmLines=0, i=0;
int shmHeadId= -1, shmDataId= -1, msgId= -1;
key_t shmHeadKey= -1;
struct shmid_ds shmDataStat;
struct msqid_ds msgStat;
char *shmKeyFile=IFCS;
int color = -1;
time_t SendFileStart=0, RecvFileStart=0, SesStart_=0;

int main (int argc, char *argv[])
{
Win wemsi, wtrans, wlog;	// EMSI, Transfer and Log windows
char key=' ';			// User key ('Q' - quit, 'TAB' - next line,
                                // 'H' - hang up)
int cols=0, lines=0;		// Terminal window size
struct tm *TM;			// Local time struct
time_t Time=0, idle=0, SesStart=0;		// Cur time, idle start time
// char speedb[8+1];		// Temp buffer for speed str
// int ttybits, prev_ttybits;	// tty state: current and previos
// struct serial_stats_struct ttyrxtx;	// tty current statistics struct
// struct serial_stats_struct prev_ttyrxtx;// tty previos statistics struct
// short CD=0, TX=0, RX=0, TR=0, hasttystat=0; // LEDS and switches
int ttydsc = -1;		// tty file descriptor
int i = 0;

if((argc-1))
    for(i = 0; i<argc; i++)
	{
	if(!strcasecmp(argv[i], "-b")) color = 0;
	}

if (readconfig())
    {
    fprintf(stderr,"Error getting configuration, aborting\n");
    exit(CONFIG_ERROR);
    }

if(ipcfile) shmKeyFile = ipcfile;

if ((shmHeadKey = ftok (shmKeyFile, 0)) < 0)
    {
    fprintf (stderr, "Unable to find shm keyfile \"%s\" (%s)\n", shmKeyFile,
							    strerror(errno));
    exit(SHM_ERROR);
    }
    else 
noifcico:
	if ((shmHeadId = shmget (shmHeadKey, sizeof(shmHead_), 0)) < 0)
	    {
	    if(errno != ENOENT)
		{
		fprintf (stderr, "%sUnable to get shm sockets segment (%s)\n",
						i ? "\n": "", strerror(errno));
		exit(SHM_ERROR);
		}
	    else
		{
		sleep(1);
		goto noifcico;
		}
	    }
	    else
		{
		if(i) fputs("", stderr);
		if((shmHead = shmat(shmHeadId, 0, 0)) < 0)
		    {
		    fprintf(stderr, "Unable attach to shm sockets segment (%s)\n",
							    strerror(errno));
		    exit(SHM_ERROR);
		    }
		}

atexit(AtExit);

if (!initscr())
    {
    fprintf (stderr, "Can't init screen\n");
    exit(NCURSES_ERROR);
    }
    
cols  = COLS;
lines = LINES;

if((cols<80) || (lines<24))
    {
    fprintf (stderr, "Need 80 colons and 24 lines at least.\n");
    exit(NCURSES_ERROR);
    }

if(savetty() == ERR) fputs ("savetty() error", stderr);
if(cbreak() == ERR) fputs ("cbreak() error", stderr);
if(noecho() == ERR) fputs ("noecho() error", stderr);
if(halfdelay(2) == ERR) fputs ("halfdelay() error", stderr);
if(curs_set(0) == ERR) fputs ("curs_set() error", stderr);
if(leaveok(stdscr, TRUE)==ERR) fputs ("leaveok() error", stderr);

if(color < 0) color=has_colors();

if (color)
    if(start_color() != ERR)
	{
	init_pair (1, COLOR_WHITE, COLOR_BLACK);
	init_pair (2, COLOR_BLACK, COLOR_WHITE);
	init_pair (3, COLOR_BLUE, COLOR_WHITE);
	init_pair (4, COLOR_RED, COLOR_WHITE);
	init_pair (5, COLOR_CYAN, COLOR_BLACK);
	init_pair (6, COLOR_YELLOW, COLOR_BLACK);
	init_pair (7, COLOR_RED, COLOR_BLACK);
	init_pair (8, COLOR_BLUE, COLOR_BLACK);
	init_pair (9, COLOR_BLACK, COLOR_RED);
	init_pair (10, COLOR_GREEN, COLOR_BLACK);
	attrset(COLOR_PAIR (1));
//	attron(COLOR_PAIR(1));
//	erase();
	}
    else
	{
	fputs ("color_start() error", stderr);
	color=FALSE;
	}

wemsi.cols = cols - 2;
wemsi.rows = 6;
wemsi.x=1;
wemsi.y=lines-1-wemsi.rows;
if(!(wemsi.w = newwin (wemsi.rows, wemsi.cols, wemsi.y, wemsi.x)))
    fputs("Cant init EMSI window", stderr);
if(wemsi.w)
    {
    leaveok(wemsi.w, TRUE);
//    wbkgd(wemsi.w, COLOR_PAIR(1) | ' ');
    }

wtrans.cols = cols-2;
wtrans.rows = 6;
wtrans.x = 1;
wtrans.y = wemsi.y-wtrans.rows;
if(!(wtrans.w = newwin (wtrans.rows, wtrans.cols, wtrans.y, wtrans.x)))
    fputs("Cant init Transfer window", stderr);
if(wtrans.w)
    {
    leaveok(wtrans.w, TRUE);
//    wbkgd(wtrans.w, COLOR_PAIR(1) | ' ');
    }
    
wlog.cols = cols - 2;
wlog.rows = lines-1-wemsi.rows-wtrans.rows-1;
wlog.x=1;
wlog.y=wtrans.y-wlog.rows;
if(!(wlog.w = newwin (wlog.rows, wlog.cols, wlog.y, wlog.x)))
    fputs("Cant init Log window", stderr);
if(wlog.w) scrollok(wlog.w, TRUE);
if(wlog.w)
    {
    leaveok(wlog.w, TRUE);
//    wbkgd(wlog.w, COLOR_PAIR(1) | ' ');
    }

if(color) attron(COLOR_PAIR(8));
border(0,0,0,0,0,0,0,0);

if(color) attron(COLOR_PAIR(2));
mvprintw (0, 2, " idle ");

while (((key = getch ()) != 'q') && (key != 'Q'))
    {
    Time=time(NULL);
    TM=localtime(&Time);
    if(color) attron(COLOR_PAIR(3));
    mvprintw (0, cols-2-5, "%02u%c%02u", TM->tm_hour,
	    ((TM->tm_sec % 2) || (!TM->tm_sec)) ? ':' : ' ', TM->tm_min);

    if (!shmLine || !shmData)
	{
badline:
	if(!idle) time(&idle);
	for(shmLine++; shmHead->Key[shmLine] < 0; shmLine++)
	    {
	    Time=time(NULL);
	    TM=localtime(&Time);

	    if((Time-idle)/60)
		{
		if(color) attron(COLOR_PAIR(3));
		mvprintw (lines-1, 2, " Idle %u min. ", (Time-idle)/60);
		}
		
	    refresh();

	    if (shmLine == UCHAR_MAX)
		{
		sleep(1);
		shmLine = 0;
		break;
		}
	    }
	if(shmLine)
	    {
	    if((shmDataId = shmget (shmHead->Key[shmLine], sizeof(shmData_), 0)) < 0)
		goto badline;
	    else 
		if((shmData = shmat(shmDataId, 0, 0)) < 0)
		    {
		    shmData = NULL;
		    goto badline;
		    }
		else
		    {
		    idle=0;
		    if((msgId = msgget(shmHead->Key[shmLine], 0))<0)
			fprintf(stderr, "Unable attach to Log msg queue (%s)",
							    strerror(errno));
		    if(color) attron(COLOR_PAIR(1));
		    border(0,0,0,0,0,0,0,0);
		    
		    if(color) attron(COLOR_PAIR(2));
		    mvprintw (0, 2, " %s, %s/%u ", shmData->Whoami,
			    shmData->SesMaster ? "Master" : "Slave", shmLine);
		    mvprintw (0, cols-2-5-2-strlen(shmData->Ver)-7, "ifcico %s",
								shmData->Ver);
		    SesStart=0;
		    SesStart_=0;
		    SendFileStart=0;
		    RecvFileStart=0;
		    if(ttydsc != -1)
			{
			close(ttydsc);
			ttydsc = -1;
			}
		    }
	    }
	}
	else
	    if ((shmData && (kill (shmData->Pid, 0) < 0)) || (key == '\t'))
		{
		werase(wlog.w);
		werase(wemsi.w);
		werase(wtrans.w);
		erase();
		refresh();
		
		if(ttydsc != -1) { close(ttydsc); ttydsc=-1; }
		
		if(key != '\t')
		    {
		    if(color) attron(COLOR_PAIR(8));
		    border(0,0,0,0,0,0,0,0);
		    
		    if(color) attron(COLOR_PAIR(2));
		    mvprintw (0, 2, " idle ");
		    }

		shmDataStat.shm_nattch = 0;
		shmctl(shmDataId, IPC_STAT, &shmDataStat);
		shmdt ((char *) shmData);
		if(shmDataStat.shm_nattch == 1) shmHead->Key[shmLine] = -1;
		shmData = NULL;
		if(ttydsc != -1)
		    {
		    close(ttydsc);
		    ttydsc=-1;
		    }
		goto badline;
		}
	    else
		{
		/*shmLines = 0;
		for(i = 1; i < UCHAR_MAX; i++)
		    {
		    if(!(shmHead->Key[i] < 0)) shmLines++;
		    }*/
		    
		if ((key == 'h') || (key == 'H'))
		    kill(shmData->Pid, SIGHUP);

		if(!(msgId<0) && wlog.w) PrintLog(wlog);
		    
		if(SesStart != shmData->SesStart)
		    {
		    SesStart=shmData->SesStart;
		    
		    werase(wemsi.w);
		    werase(wtrans.w);
		    
		    wrefresh(wlog.w);
		    wrefresh(wemsi.w);
		    wrefresh(wtrans.w);
		    
		    if(color) attron(COLOR_PAIR(1));
		    mvhline(lines-1, 1, ACS_HLINE, cols-2);
		    
		    if(SesStart && wemsi.w) PrintEMSIData(wemsi);
		    }
		    
		if(color) attron(COLOR_PAIR(2));
/*		
		if(strlen(shmData->tty))
		    {
		    if(ttydsc == -1) ttydsc=open(shmData->tty, O_NDELAY | O_RDONLY);
		    
		    if(ttydsc != -1)
			{
			hasttystat=1;
			
			if(ioctl(ttydsc, TIOCSERGSTATS, &ttyrxtx) != -1)
			    {
			    if((ttyrxtx.rx > prev_ttyrxtx.rx) && (ttyrxtx.count > 1))
				RX=TRUE;
			    else RX=FALSE;
			    
			    if((ttyrxtx.tx > prev_ttyrxtx.tx) && (ttyrxtx.count > 1))
				TX=TRUE;
			    else TX=FALSE;
			    
			    prev_ttyrxtx = ttyrxtx;
			    }
			else hasttystat=0;
			
			if((hasttystat)&&(ioctl(ttydsc,TIOCMGET,&ttybits) != -1))
			    {
			    if(ttybits & TIOCM_DTR) TR=TRUE;
			    else TR=FALSE;
			    
			    if(ttybits & TIOCM_CAR) CD=TRUE;
			    else CD=FALSE;
			    
			    prev_ttybits = ttybits;
			    }
			else hasttystat=0;
			
			if(hasttystat)
			    {
			    // "TR CD RX TX "
			    if(color) attron(COLOR_PAIR(2));
			    mvprintw (lines-1, cols-3, " ");
			
			    if(color && RX) attron(COLOR_PAIR(9));
			    mvprintw (lines-1, cols-3-2, "RX");
			
			    if(color) attron(COLOR_PAIR(2));
			    mvprintw (lines-1, cols-3-3, " ");
			
			    if(color && TX) attron(COLOR_PAIR(9));
			    mvprintw (lines-1, cols-3-5, "TX");
			
			    if(color) attron(COLOR_PAIR(2));
			    mvprintw (lines-1, cols-3-6, " ");
			
			    if(color && CD) attron(COLOR_PAIR(9));
			    mvprintw (lines-1, cols-3-8, "CD");
			
			    if(color) attron(COLOR_PAIR(2));
			    mvprintw (lines-1, cols-3-9, " ");
			
			    if(color && TR) attron(COLOR_PAIR(9));
			    mvprintw (lines-1, cols-3-11, "TR");
			    }
			}
		    else hasttystat=0;
		    
		    if(color) attron(COLOR_PAIR(2));
		    if(shmData->ttyspeed)
		        {
		        snprintf(speedb, 7, "%u", shmData->ttyspeed);
			mvprintw (lines-1, cols-strlen(shmData->tty)-5-strlen(speedb)-
			    (12*hasttystat), " %s %s ", shmData->tty, speedb);
			}
		    else
		        mvprintw (lines-1, cols-strlen(shmData->tty)-4-(12*hasttystat),
			                             " %s ", shmData->tty);
		    }
		else
		    {
		    if(ttydsc != -1)
			{
			close(ttydsc);
			ttydsc = -1;
			}
		    if(color) attron(COLOR_PAIR(1));
		    mvhline(lines-1, 1, ACS_HLINE, cols-2);
		    }
*/
		if(SesStart && ((Time-SesStart)/60))
		    {
		    if(color) attron(COLOR_PAIR(3));
		    mvprintw (lines-1, 2, " Online %u min. ", (Time-SesStart)/60);
		    }

		refresh();

		if(shmData->SesStart && wtrans.w) PrintTransfer(wtrans);
		}
    }
    
exit(NORMAL);
}


void PrintLog (Win w)
{
char *p;
struct msg_
    {
    long mtype;
    char mtext[w.cols * 2 + 1 - 6];
    } msg;

if(msgrcv(msgId, &msg, sizeof(msg.mtext) - 1, 0, IPC_NOWAIT | MSG_NOERROR)<0)
    return;
else
    {
    scroll(w.w);
    if(color) wattron(w.w, COLOR_PAIR(msg.mtype));
    if(strlen(msg.mtext) > w.cols)
	{
	msg.mtext[w.cols * 2] = '\0';
	p = strdup(msg.mtext + w.cols);
	msg.mtext[w.cols] = '\0';
	mvwprintw(w.w, w.rows - 1, 0, msg.mtext);
	mvwprintw(w.w, w.rows - 1, 5, p);
	free(p);
	}
    else mvwprintw(w.w, w.rows-1, 0, msg.mtext);
    wrefresh (w.w);
    }
}


void PrintTransfer (Win w)
{
const char *STC_header = {" Size   Time  CPS "};
const char *ses_header = {"  Total  Session  "};
const char *cur_header = {"   Current file   "};

char Buff[w.cols+1];

struct tm *TM;
time_t ttmp;
static prev=0;

struct TrState
    {
    unsigned long TrSize, TrPos, TrSkip, TrCPS, TrToGo, Pos, CPS, ToGo, Size, Skip;
    unsigned short TrToGoPers, ToGoPers;
    time_t TrTime, TrToGoTime, Time, ToGoTime;
    } Send;
    
struct TrState Recv;

if(prev < time(NULL)) prev=time(NULL);
else return;

if(SesStart_ != shmData->SesStart)
    {
    SesStart_ = shmData->SesStart;
    
    if(color) wattron(w.w, COLOR_PAIR(1));
    mvwhline(w.w, 5, 0, ACS_HLINE, w.cols);
    mvwhline(w.w, 0, 0, ACS_HLINE, w.cols);
        
    if(color) wattron(w.w, COLOR_PAIR(6));
    if(strlen(shmData->OtherEndPhone))
	{
	if(shmData->OtherEndLines>1)
	mvwprintw(w.w, 5, w.cols/2+1, " %s [%u] ",
		shmData->OtherEndPhone, shmData->OtherEndLine+1);
	else
	    mvwprintw(w.w, 5, w.cols/2+1, " %s ", shmData->OtherEndPhone);
	}
	
    if(color) wattron(w.w, COLOR_PAIR(4));
    mvwprintw (w.w, 1, 1, STC_header);
    mvwprintw (w.w, 1, 1+strlen(STC_header)+1, STC_header);
    mvwprintw (w.w, 1, w.cols/2+1, STC_header);
    mvwprintw (w.w, 1, w.cols/2+1+strlen(STC_header)+1, STC_header);
    mvwprintw (w.w, 3, 1, ses_header);
    mvwprintw (w.w, 3, 1+strlen(ses_header)+1, cur_header);
    mvwprintw (w.w, 3, w.cols/2+1, ses_header);
    mvwprintw (w.w, 3, w.cols/2+1+strlen(ses_header)+1, cur_header);
    }


if(SendFileStart != shmData->send.Start)
    {
    SendFileStart = shmData->send.Start;
    
    if(color) wattron(w.w, COLOR_PAIR(1));
    mvwhline(w.w, 0, 0, ACS_HLINE, w.cols/2);
    
    strncpy(Buff, shmData->send.FileName, (w.cols-8)/2);
    Buff[(w.cols-8)/2]='\0';
    
    if(color) wattron(w.w, COLOR_PAIR(6));
    if(strlen(Buff)) mvwprintw(w.w, 0, (w.cols/2-2-strlen(Buff))/2, " %s ", Buff);
    }

if(RecvFileStart != shmData->recv.Start)
    {
    RecvFileStart = shmData->recv.Start;

    if(color) wattron(w.w, COLOR_PAIR(1));
    mvwhline(w.w, 0, w.cols/2, ACS_HLINE, w.cols/2);

    strncpy(Buff, shmData->recv.FileName, (w.cols-8)/2);
    Buff[(w.cols-8)/2]='\0';
    
    if(color) wattron(w.w, COLOR_PAIR(6));
    if(strlen(Buff)) mvwprintw(w.w, 0, w.cols/2+((w.cols/2-2-strlen(Buff))/2), " %s ", Buff);
    }

if(shmData->EMSI_Time)
    {
    ttmp=shmData->EMSI_Time + difftime(time(NULL), shmData->SesStart);
    TM=localtime(&ttmp);

    if(color) wattron(w.w, COLOR_PAIR(3));
    mvwprintw(w.w, 5, w.cols-1-5, "%02u%c%02u", TM->tm_hour,
			((TM->tm_sec % 2) || (!TM->tm_sec)) ? ':' : ' ',
						TM->tm_min);
    }

if(shmData->send.BlockSize)
    {
    if(color) wattron(w.w, COLOR_PAIR(5));
    mvwprintw(w.w, 5, w.cols/2-1-5-1-6, "[%4u]", shmData->send.BlockSize);
    }
else
    {
    if(color) wattron(w.w, COLOR_PAIR(1));
    mvwhline(w.w, 5, w.cols/2-1-5-1-6, ACS_HLINE, 6);
    }
    
if(shmData->recv.BlockSize)
    {
    if(color) wattron(w.w, COLOR_PAIR(5));
    mvwprintw(w.w, 5, w.cols-1-5-1-6, "[%4u]", shmData->recv.BlockSize);
    }
else
    {
    if(color) wattron(w.w, COLOR_PAIR(1));
    mvwhline(w.w, 5, w.cols-1-5-1-6, ACS_HLINE, 6);
    }

Send.TrSize=shmData->send.TrSize;
Recv.TrSize=shmData->recv.TrSize;

Send.TrSkip=shmData->send.TrSkip;
Recv.TrSkip=shmData->recv.TrSkip;

Send.Pos=shmData->send.Pos;
Recv.Pos=shmData->recv.Pos;

Send.Size=shmData->send.FileSize;
Recv.Size=shmData->recv.FileSize;

Send.Skip=shmData->send.SkipSize;
Recv.Skip=shmData->recv.SkipSize;

if(Send.Pos>=Send.Skip) Send.TrPos=shmData->send.TrFiles+Send.Pos-Send.Skip;
else Send.TrPos=0;
if(Recv.Pos>=Recv.Skip) Recv.TrPos=shmData->recv.TrFiles+Recv.Pos-Recv.Skip;
else Recv.TrPos=0;

if(shmData->send.Start) Send.Time = difftime (time(NULL), shmData->send.Start);
else Send.Time=0;
if(shmData->recv.Start) Recv.Time = difftime (time(NULL), shmData->recv.Start);
else Recv.Time=0;

if (Send.Time && (Send.Pos >= Send.Skip))
    Send.CPS=(Send.Pos-Send.Skip)/Send.Time;
else
    Send.CPS=0L;
    
if (Recv.Time && (Recv.Pos >= Recv.Skip))
    Recv.CPS=(Recv.Pos-Recv.Skip)/Recv.Time;
else
    Recv.CPS=0L;

if(shmData->recv.TrStart && shmData->recv.Start)
    Recv.TrTime = difftime (time(NULL), shmData->recv.TrStart);
else
    Recv.TrTime=0;

if(shmData->send.TrStart && shmData->send.Start)
    Send.TrTime = difftime (time(NULL), shmData->send.TrStart);
else
    Send.TrTime=0;

if (Recv.TrTime && Recv.TrPos)
    Recv.TrCPS = Recv.TrPos/Recv.TrTime;
else
    Recv.TrCPS=0;
    
if (Send.TrTime && Send.TrPos)
    Send.TrCPS = Send.TrPos/Send.TrTime;
else
    Send.TrCPS=0;

if(Send.TrSize>=(Send.TrSkip+Send.TrPos))
    Send.TrToGo=Send.TrSize-Send.TrSkip-Send.TrPos;
else
    Send.TrToGo=0L;

if(Recv.TrSize>=(Recv.TrSkip+Recv.TrPos))
    Recv.TrToGo=Recv.TrSize-Recv.TrSkip-Recv.TrPos;
else
    Recv.TrToGo=0L;

if(Send.TrSize>=100)
    {
    if((Send.TrToGoPers=Send.TrToGo/(Send.TrSize/100))>100) Send.TrToGoPers=100;
    }
else Send.TrToGoPers=0;
if(Recv.TrSize>=100)
    {
    if((Recv.TrToGoPers=Recv.TrToGo/(Recv.TrSize/100))>100) Recv.TrToGoPers=100;
    }
else Recv.TrToGoPers=0;

if(Send.Size>=Send.Pos) Send.ToGo=Send.Size-Send.Pos;
else Send.ToGo=0L;

if(Recv.Size>=Recv.Pos) Recv.ToGo=Recv.Size-Recv.Pos;
else Recv.ToGo=0L;

if(Send.Size>=100)
    {
    if((Send.ToGoPers=Send.ToGo/(Send.Size/100))>100) Send.ToGoPers=100;
    }
else Send.ToGoPers=0;
if(Recv.Size>=100)
    {
    if((Recv.ToGoPers=Recv.ToGo/(Recv.Size/100))>100) Recv.ToGoPers=100;
    }
else Recv.ToGoPers=0;

if(Send.TrCPS) Send.TrToGoTime=Send.TrToGo/Send.TrCPS;
else Send.TrToGoTime=0;
if(Recv.TrCPS) Recv.TrToGoTime=Recv.TrToGo/Recv.TrCPS;
else Recv.TrToGoTime=0;

if(Send.CPS) Send.ToGoTime=Send.ToGo/Send.CPS;
else Send.ToGoTime=0;
if(Recv.CPS) Recv.ToGoTime=Recv.ToGo/Recv.CPS;
else Recv.ToGoTime=0;

if(shmData->send.Start)
    {
    if(color) wattron(w.w, COLOR_PAIR(5));
    }
else
    if(color) wattron(w.w, COLOR_PAIR(8));

mvwprintw (w.w, 2, 0, "%6lu%s %2u:%02u %4lu%s",
		Send.TrPos/1024 ? Send.TrPos/1024 : Send.TrPos,
		Send.TrPos/1024 ? "k" : " ",
		Send.TrTime>=(60*60) ? Send.TrTime/(60*60) : Send.TrTime/60,
		Send.TrTime>=(60*60) ? (Send.TrTime%(60*60)/60):Send.TrTime%60,
		Send.TrCPS>9999 ? Send.TrCPS/1024 : Send.TrCPS,
		Send.TrCPS>9999 ? "k" : " ");
mvwprintw (w.w, 4, 0, "%6lu%s %2u:%02u  %3lu%%",
		Send.TrToGo/1024 ? Send.TrToGo/1024 : Send.TrToGo,
		Send.TrToGo/1024 ? "k" : " ",
		Send.TrToGoTime>=(60*60)?Send.TrToGoTime/(60*60):Send.TrToGoTime/60,
		Send.TrToGoTime>=(60*60)?(Send.TrToGoTime%(60*60))/60:Send.TrToGoTime%60,
		Send.TrToGoPers);
mvwprintw (w.w, 2, 6+1+1+2+1+2+1+3+1+1, "%6lu%s %2u:%02u %4lu%s",
		Send.Pos/1024 ? Send.Pos/1024 : Send.Pos,
		Send.Pos/1024 ? "k" : " ",
		Send.Time>=(60*60)?Send.Time/(60*60):Send.Time/60,
		Send.Time>=(60*60)?(Send.Time%(60*60))/60:Send.Time%60,
		Send.CPS>9999 ? Send.CPS/1024 : Send.CPS,
		Send.CPS>9999 ? "k" : " ");
mvwprintw (w.w, 4, 6+1+1+2+1+2+1+3+1+1, "%6lu%s %2u:%02u  %3lu%%",
		Send.ToGo/1024 ? Send.ToGo/1024 : Send.ToGo,
		Send.ToGo/1024 ? "k" : " ",
		Send.ToGoTime>=(60*60)?Send.ToGoTime/(60*60):Send.ToGoTime/60,
		Send.ToGoTime>=(60*60)?(Send.ToGoTime%(60*60))/60:Send.ToGoTime%60,
		Send.ToGoPers);

if(Recv.TrSize)
    {
    if(color) wattron(w.w, COLOR_PAIR(5));
    }
else
    if(color) wattron(w.w, COLOR_PAIR(8));
mvwprintw (w.w, 4, w.cols/2, "%6lu%s %2u:%02u  %3lu%%",
		Recv.TrToGo/1024 ? Recv.TrToGo/1024 : Recv.TrToGo,
		Recv.TrToGo/1024 ? "k" : " ",
		Recv.TrToGoTime>=(60*60)?Recv.TrToGoTime/(60*60):Recv.TrToGoTime/60,
		Recv.TrToGoTime>=(60*60)?(Recv.TrToGoTime%(60*60))/60:Recv.TrToGoTime%60,
		Recv.TrToGoPers);

if(shmData->recv.Start)
    {
    if(color) wattron(w.w, COLOR_PAIR(5));
    }
else
    if(color) wattron(w.w, COLOR_PAIR(8));
mvwprintw (w.w, 2, w.cols/2, "%6lu%s %2u:%02u %4lu%s",
		Recv.TrPos/1024 ? Recv.TrPos/1024 : Recv.TrPos,
		Recv.TrPos/1024 ? "k" : " ",
		Recv.TrTime>=(60*60) ? Recv.TrTime/(60*60) : Recv.TrTime/60,
		Recv.TrTime>=(60*60) ? (Recv.TrTime%(60*60))/60:Recv.TrTime%60,
		Recv.TrCPS>9999 ? Recv.TrCPS/1024 : Recv.TrCPS,
		Recv.TrCPS>9999 ? "k" : " ");
mvwprintw (w.w, 2, w.cols/2+6+1+1+2+1+2+1+3+1+1, "%6lu%s %2u:%02u %4lu%s",
		Recv.Pos/1024 ? Recv.Pos/1024 : Recv.Pos,
		Recv.Pos/1024 ? "k" : " ",
		Recv.Time>=(60*60) ? Recv.Time/(60*60) : Recv.Time/60,
		Recv.Time>=(60*60) ? (Recv.Time%(60*60))/60 : Recv.Time%60,
		Recv.CPS>9999 ? Recv.CPS/1024 : Recv.CPS,
		Recv.CPS>9999 ? "k" : " ");
mvwprintw (w.w, 4, w.cols/2+6+1+1+2+1+2+1+3+1+1, "%6lu%s %2u:%02u  %3lu%%",
		Recv.ToGo/1024 ? Recv.ToGo/1024 : Recv.ToGo,
		Recv.ToGo/1024 ? "k" : " ",
		Recv.ToGoTime>=(60*60)?Recv.ToGoTime/(60*60):Recv.ToGoTime/60,
		Recv.ToGoTime>=(60*60)?(Recv.ToGoTime%(60*60))/60:Recv.ToGoTime%60,
		Recv.ToGoPers);

wrefresh (w.w);
}

void PrintEMSIData (Win w)
{
int y=0, x;
char Buff[w.cols+1], *protected_msg = {"* Password protected *"};

snprintf(Buff,(size_t) w.cols, "* %s call%s", shmData->SesMaster ?
	    "Outgoing" : "Incoming", strlen(shmData->ConnectStr) ? ": " : "");

if(strlen(Buff))
    {
    x=(w.cols-strlen(Buff)-2-strlen(shmData->ConnectStr))/2;
    if(color) wattron(w.w, COLOR_PAIR(1));
	mvwprintw (w.w, y, x, Buff);
    if(color) wattron(w.w, COLOR_PAIR(6));
	mvwprintw (w.w, y, x+strlen(Buff), shmData->ConnectStr);
    if(color) wattron(w.w, COLOR_PAIR(1));
	mvwprintw (w.w, y++, x+strlen(Buff)+strlen(shmData->ConnectStr), " *");
    }
	
if(color) wattron(w.w, COLOR_PAIR(5));
    
if(strlen(shmData->EMSI_Addr) || strlen(shmData->EMSI_Using))
    {
    snprintf(Buff,(size_t) w.cols, "%s%s%s", shmData->EMSI_Addr, 
			strlen(shmData->EMSI_Using) ? ", using " : "", shmData->EMSI_Using);
    if(strlen(Buff))
        mvwprintw (w.w, y++, (w.cols-strlen(Buff))/2, Buff);
    }

if(strlen(shmData->EMSI_System) || strlen(shmData->EMSI_Location) || 
		    strlen(shmData->EMSI_SySop) || strlen(shmData->EMSI_Phone))
    {
    snprintf(Buff,(size_t) w.cols, "%s%s%s%s%s%s%s", shmData->EMSI_System, 
        strlen(shmData->EMSI_Location) ? ", " : "", shmData->EMSI_Location,
        strlen(shmData->EMSI_SySop) ? ", " : "", shmData->EMSI_SySop,
        strlen(shmData->EMSI_Phone) ? ", " : "", shmData->EMSI_Phone);
    if(strlen(Buff))
        mvwprintw (w.w, y++, (w.cols-strlen(Buff))/2, Buff);
    }
    
if(strlen(shmData->EMSI_Flags))
    {
    strncpy(Buff, shmData->EMSI_Flags, (size_t) w.cols);
    if(strlen(Buff))
        mvwprintw (w.w, y++, (w.cols-strlen(Buff))/2, Buff);
    }
    
if (strlen(shmData->EMSI_AKA))
    {
    snprintf(Buff,(size_t) w.cols, "AKA %s", shmData->EMSI_AKA);
    if(strlen(Buff))
        mvwprintw (w.w, y++, (w.cols-strlen(Buff))/2, Buff);
    }
    
if(color) wattron(w.w, COLOR_PAIR(7));
if (shmData->SesProtected)
    mvwprintw (w.w, y++, (w.cols-strlen(protected_msg))/2, protected_msg);
    
wrefresh (w.w);
}

void usage(void)
{
confusage("");
}

void AtExit(void)
{
erase();
refresh();
nocbreak ();
resetty ();
endwin ();

if(shmData)
    {
    if(shmdt (shmData) < 0)
	{
	printf("Error detach from shm segment (%s)", strerror(errno));
	exit(SHM_ERROR);
	}
    }
if(shmdt (shmHead) < 0)
    {
    printf("Error detach from shm sockets segment (%s)", strerror(errno));
    exit(SHM_ERROR);
    }
}
#else
void main() {}
#endif
