/*
     Janus (full-duplex bidirectional protocol) implementation for ifcico mailer
     Copyright (C) 1999	Kolya Nesterov, 2:463/567@fidonet
 
     This part of the program is free software; you can redistribute it 
     and/or modify it under the terms of the GNU General Public License
     as published by the Free Software Foundation; either version 2 of 
     the License, or (at your option) any later version.
 
     This part of the program is distributed in the hope that it will be 
     useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.
 
     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
     Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
/* TODO: See code for notes. Code refinement. Testing, testing, testing ....
 * Variable BLK sizes.
 * Maybe it would be better to rewrite it using posix threads ?
 * Code will be less complicated and more bug-protected
 * 
 */
#include <stdlib.h>

#include <stdio.h>
#include <string.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>
 
#include <errno.h>
#include <time.h>
#include <assert.h>
#include <crc.h>
#include "fsm.h"
#include "xutil.h"
#include "lutil.h"
#include "ttyio.h"
#include "session.h"
#include "config.h"
#include "emsi.h"
#include "janus.h"
#include "openfile.h"
#ifdef IFCS
#include "ifcs.h"
void ipcloginf(int, char*,...);
extern shmData_ *shmData;
#endif

/* max data block size */
#define JBLKSIZE 2048
/* 2048 plus 4 bytes of offset */
#define JTXBUFSIZE (JBLKSIZE + 4) 
/* data + type + 2 bytes of possible garbage */
#define JRXBUFSIZE (JTXBUFSIZE + 3)

/* max resulting pkt size. as twice as real non-escaped packet */
#define PKTBUFSIZE ((2 + JTXBUFSIZE + 1 + 2 + 4) * 2)

#define DEAD_TIMER	0
#define RPOS_TIMER	1
#define ACK_TIMER	2

#define  ACK_TIMEOUT	30
#define RECV_TIMEOUT	10
#define RPOS_TIMEOUT	40
#define DEAD_TIMEOUT	120
#define MAX_ACK_COUNT	5

/* FIXME: redefine this to support different sizes and byte orders */
typedef unsigned long UINT32;
typedef unsigned short UINT16;
typedef unsigned char UCHAR;
#define MACHINE_TO_LSB_32(a) (a)
#define MACHINE_TO_LSB_16(a) (a)
#define LSB_TO_MACHINE_32(a) (a)
#define LSB_TO_MACHINE_16(a) (a)

#define TRUE (1==1)
#define FALSE (1==0)

file_list *respond_bark(char *buf);
extern char *reqname(faddr*);
/*Internal janus variables and functions */
static char *pkt_buffer = NULL;

static int sendpkt(int, int,char*,int);
extern FILE *flockopen(const char *fname, const char *mode);
static void escape_byte(int clear8, char b, char **p);
static int get_unescaped_byte(int timeout);
static int rcvpkt(int *size, char *buf, int timeout);
static int end_janus_session(int haltreceived);
static char *read_freqs(fa_list *remote);
/* Just for simplifying the code 
 * Using the functions is not suitable here, IMHO
 */
#define LOGFILE(type, msg)\
	loginf("janus: %s %lu bytes in %.0lf seconds (%.0lf cps)", msg, \
			(unsigned long) type ## size - type ## startsize, \
			difftime(time(NULL), type ## time),\
			(type ## size - type ## startsize) / \
			difftime(time(NULL), type ## time));

int janus()
{
BSM_NAME("janus", 1)
BSM_STATES
	xsendfname, xrcvfnack, xsendblk, xrcveofack, xrcvrpos, xwaitpkt,
	xrcvfreq, xrcvfrqnakack, xsendfreqnak,
	rrcvfname, rrcvblk, rrcvfreqnak, rwaitpkt
BSM_EDECL
	char  *txbuf,		*rxbuf;
	size_t txbsize,		rxbsize, txblksize, sentbytes, blkstepbytes;
	time_t txtime,		rxtime,	lastrpostime;
	FILE  *txfp,		*rxfp;
	off_t  txstartsize, txsize, rxstartsize, rxsize;
	
	int	pkttype, remotecaps, restartstate, ackcount, rc;
	struct stat txst;
	char	*reqfiles, *reqp;
	
	/**/
	fa_list 	*eff_remote,tmpl;
	file_list	*tosend,*tmpfl=NULL;
	char *nonhold_mail;
	/**/

	RESETTIMERS();
	SETTIMER(DEAD_TIMER, DEAD_TIMEOUT);

	/* Allocate buffers, set variables */
	pkt_buffer	= xmalloc(PKTBUFSIZE);
	txbuf		= xmalloc(JTXBUFSIZE);
	rxbuf		= xmalloc(JRXBUFSIZE + sizeof(unsigned long) + sizeof(int));
	rxfp		= txfp	= NULL;
	pkttype		= JPKT_NONE;
	remotecaps	= 0;
	restartstate	= 0;
	ackcount	= 0;
	rxbsize = txbsize = rxsize = txsize = txstartsize = sentbytes = 0L;
	lastrpostime 	= (time_t) 0L;
	txblksize	= JBLKSIZE;
	blkstepbytes	= 0L;
	rc		= 0;
	
	loginf("janus: start transfer");
#ifdef IFCS
	  ipcloginf (1, "janus: start transfer");
#endif
	session_flags |= SESSION_JANUS;
	
	/* Now create filelist of files to be sent 
	 * TODO: Move this to a better place
	 */
	/**/
	if (emsi_remote_lcodes & LCODE_NPU)
	{
		loginf("remote requested \"no pickup\", no send");
#ifdef IFCS
		ipcloginf(8, "janus: remote requested \"no pickup\", no send");
#endif
		eff_remote=NULL;
	}
	else if (emsi_remote_lcodes & LCODE_PUP)
	{
		loginf("remote requested \"pickup primary\"");
#ifdef IFCS
		ipcloginf(8, "janus: remote requested \"pickup primary\"");
#endif
		tmpl.addr=remote->addr;
		tmpl.next=NULL;
		eff_remote=&tmpl;
	}
	else eff_remote=remote;
	if (localoptions & NOHOLD) nonhold_mail=ALL_MAIL;
	else nonhold_mail=NONHOLD_MAIL;
	if (emsi_remote_lcodes & LCODE_HAT) {
 		loginf("remote asked to \"hold all traffic\", no send");
#ifdef IFCS
		ipcloginf(8, "janus: remote asked to \"hold all traffic\", no send");
#endif
		tosend = NULL;
	} else {
		tmpfl=tosend = create_filelist(eff_remote,nonhold_mail,0);
	}
	reqp = reqfiles = read_freqs(eff_remote);
	/**/
	

BSM_START(xsendfname, rwaitpkt)

TRX_STATE(xwaitpkt, rwaitpkt)
/* ----------------------------------------------------------------------------
 * Wait for packets and dispose them
 * ----------------------------------------------------------------------------
 */
	/* I don't like how this part is done */

	if (ackcount > MAX_ACK_COUNT) {
		logerr("janus: too much ACK retries");
#ifdef IFCS
		ipcloginf (7, "janus: too much ACK retries");
#endif
		BSM_ERROR;
	};
	if (pkttype != JPKT_NONE) rxbsize = 0; /* clear the size if any packet was received before */
	pkttype = rcvpkt(&rxbsize, rxbuf, tx_state == xsendfname || 
			 tx_state == xsendblk  ? 0 : RECV_TIMEOUT);
	switch (pkttype) {
		case JPKT_BAD:	sentbytes = 0;
		case JPKT_NONE:	
				/* Check if remote is dead already */
				/* Warning: and what if we are only transmitting ? */
				if (RUNNING(DEAD_TIMER)) {
					if (EXPIRED(DEAD_TIMER)) {
						logerr("janus: global timeout expiried, aborting...");
#ifdef IFCS
						ipcloginf (7, "janus: global timeout expiried, aborting...");
#endif
						BSM_ERROR;
					};
				} else {
					SETTIMER(DEAD_TIMER, DEAD_TIMEOUT);
				};
				if (RUNNING(ACK_TIMER) && EXPIRED(ACK_TIMER)) {
					logerr("janus: timeout waiting for ACK");
#ifdef IFCS
					ipcloginf (7, "janus: timeout waiting for ACK");
#endif
					TX_PROCEED(restartstate);
					ackcount++;
				};
				break;

		case JPKT_HALT:	tx_success = rx_success = 1;
				break; /*TX_PROCEED(xrcvhalt);*/
		case JPKT_FNACK: case JPKT_RPOS:    case JPKT_EOFACK:
		case JPKT_HALTACK: case JPKT_FRNAKACK: case JPKT_FREQ: 
			/* TX packets. */
			if (!(tx_state == xwaitpkt || tx_state == xsendblk) || tx_success) {
				debug(DBG_JANUS, "xwaitpkt: pkt %c received but janus is in state %d",
						pkttype, tx_state);
				break;
			};
			RESETTIMER(DEAD_TIMER);
			RESETTIMER(ACK_TIMER);
			switch (pkttype) {
				case JPKT_FNACK:	TX_PROCEED(xrcvfnack);
				case JPKT_RPOS:		TX_PROCEED(xrcvrpos);
				case JPKT_EOFACK:	TX_PROCEED(xrcveofack);
				case JPKT_HALTACK:	break; /*TX_PROCEED(xrcvhaltack);*/
				case JPKT_FREQ:		TX_PROCEED(xrcvfreq);
				case JPKT_FRNAKACK:	TX_PROCEED(xrcvfrqnakack);
			};
			case_state = 0; /* Flip the state to handle packet received in next turn */
			break;
		case JPKT_FNAME: case JPKT_BLK: case JPKT_FREQNAK:
			/* RX packets */
			if (rx_success) {
				debug(DBG_JANUS, "rwaitpkt: pkt %c received but janus is in state %d",
						pkttype, tx_state);
				break;
			};

			RESETTIMER(DEAD_TIMER);
			switch (pkttype) {
				case JPKT_FNAME:	RX_PROCEED(rrcvfname);
				case JPKT_BLK:		RX_PROCEED(rrcvblk);
				case JPKT_FREQNAK:	RX_PROCEED(rrcvfreqnak);
			};
			case_state = 1; /* Flip the state to handle packet received in next turn */
			break;	 
		default:
			debug(DBG_JANUS, "unknown pkt type %c, size: %lu, data: %s", 
				pkttype, rxbsize, printable(rxbuf, rxbsize));
			break;
	};
	
TX_STATE(xsendfname)
/* ----------------------------------------------------------------------------
 * prepare and send FNAME packet initalizing file transfer 
 * ----------------------------------------------------------------------------
 */
	while (tmpfl) {

		if (tmpfl->remote) {
			 
			if ((txfp = flockopen(tmpfl->local, "rb")) != NULL) {
				fstat(fileno(txfp), &txst); 
				txbsize = sprintf(txbuf, "%s%c%lu %lo %o%c%c", tmpfl->remote, 
						'\0', (unsigned long) txst.st_size,
						txst.st_mtime, txst.st_mode, '\0', JCAP_OUR);
				txsize = txst.st_size;
				txstartsize = -1L;
#ifdef IFCS
				ipcloginf(6, "janus: send \"%s\", %lu bytes",
			       		S(tmpfl->remote), (unsigned long) txst.st_size);
				if(shmData) 
	    			{
		    			if(snprintf(shmData->send.FileName, PATH_MAX,
						S(tmpfl->remote))<0)
					{
						shmData->send.FileName[PATH_MAX-1]='>';
						shmData->send.FileName[PATH_MAX]='\0';
					}
		    			shmData->send.FileSize = txst.st_size;
		    			shmData->send.Start = time (NULL);
			    		if(!shmData->send.TrStart) shmData->send.TrStart = time (NULL);
		    		}
#endif		
				break;
			} else {
				debug(DBG_JANUS, "xsendfname: failed to send file %s (remote %s, errno %d)", 
						S(tmpfl->local), S(tmpfl->remote), errno);
			};
		};
		/* this has to be done to remove flo files */
		execute_disposition(tmpfl); 
		tmpfl = tmpfl->next; 
	};

	if (!tmpfl) {
		debug(DBG_JANUS, "sending NULL FNAME packet");
		memset(txbuf, '\0', txbsize = 2); /*FIXME: Does this packet works ? */
		txbuf[txbsize++] = (UCHAR )JCAP_OUR;
	};
	
	sendpkt(0, JPKT_FNAME, txbuf, txbsize + 1);
	/* Set timeout for ack */
	SETTIMER(ACK_TIMER, ACK_TIMEOUT); restartstate = xsendfname;
	TX_PROCEED(xwaitpkt);

TX_STATE(xrcvfnack)
/* ----------------------------------------------------------------------------
 * receive remote responce to FNAME packet in which it tells us about 
 * starting offset in tx file 
 * ----------------------------------------------------------------------------
 */
	if (txfp == NULL && tmpfl) {
		debug(DBG_JANUS, "xrcvfnack: error, fnack received but tx file is closed");
		TX_ERROR(xwaitpkt);
	};
	
	if (rxbsize < 4) {
		debug(DBG_JANUS, "xrcvfnack: error, fnack packet size is %ld", rxbsize);
		TX_ERROR(xwaitpkt);
	};
	
	ackcount = 0; /* We've received it */
	
	/* Get remote capabilities */
	/* WARNING: according to bterm sources it's byte,
	   but when connecting to T-mail and Bink/+ it's word (2 bytes) ! */
	
	remotecaps = rxbsize  >= 5 ? *((UCHAR *) rxbuf + sizeof(UINT32)) : 0;
	debug(DBG_JANUS, "xrcvfnack: remote caps = %x", remotecaps);
 	 
	if (!tmpfl) {
		TX_SUCCESS(xwaitpkt);
	};
		
	txstartsize = LSB_TO_MACHINE_32(*((UINT32 *) rxbuf));
	if (txstartsize != -1L) {
		/* Let's send it ! */
		fseek(txfp, txstartsize, SEEK_SET); /*FIXME: Offset check ?*/
		loginf("janus: send \"%s\" as \"%s\" (%lu bytes) starting from offset %lu",
		       S(tmpfl->local), S(tmpfl->remote), (unsigned long) txst.st_size, txstartsize);
		(void)time(&txtime);
#ifdef IFCS
		if(txstartsize>0)
		 	ipcloginf(5, "janus: sending from offset %lu", (unsigned long) txstartsize);
		if(shmData) 
    		{
			shmData->send.SkipSize = txstartsize;
			shmData->send.TrSkip += txstartsize;
		}
#endif		
		TX_PROCEED(xsendblk);
	} else {
		/* Remote asks us to skip file */
		loginf("janus: skip file \"%s\"", S(tmpfl->local));
#ifdef IFCS
		ipcloginf(8, "janus: skip file \"%s\"", S(tmpfl->local));
#endif		
		fclose(txfp); txfp = NULL;
		execute_disposition(tmpfl);
		tmpfl = tmpfl->next;
		TX_PROCEED(xsendfname);
	};

TX_STATE(xsendblk)
/* ----------------------------------------------------------------------------
 * send data block packet with yet another part of file
 * ----------------------------------------------------------------------------
 */
	
	*((UINT32 *) txbuf) = MACHINE_TO_LSB_32(ftell(txfp));
	/*TODO: implement variable size packets in a better way */
	txbsize = fread(txbuf + sizeof(UINT32), 1, txblksize, txfp);
	
	sentbytes += txbsize;
	if (sentbytes > blkstepbytes) {
	/* We sent more than needed for packet increasing ? 
	 * Let's increase block size 
	 */
		txblksize <<= 1;
		if (txblksize > JBLKSIZE) txblksize = JBLKSIZE;
		sentbytes = 0;	
	};
	
	if (ferror(txfp)) {
		logerr("janus: error reading tx file %s", S(tmpfl->local));
#ifdef IFCS
		ipcloginf (7, "janus: error reading tx file %s", S(tmpfl->local));
#endif
		TX_ERROR(xwaitpkt);
	};
	sendpkt(remotecaps, JPKT_BLK, txbuf, txbsize += 4);
#ifdef IFCS
	if(shmData) 
	{
		shmData->send.Pos = ftell(txfp);
		shmData->send.BlockSize = txbsize - 4;
	}
#endif
	if (feof(txfp)) {
		/* Set timeout for ack */
		SETTIMER(ACK_TIMER, ACK_TIMEOUT); restartstate = xsendblk;
		TX_PROCEED(xwaitpkt);
	} else { 
		RESETTIMER(DEAD_TIMER);
		TX_PROCEED(xsendblk);
	};

TX_STATE(xrcvrpos)	
/* ----------------------------------------------------------------------------
 * receive rpos packet (if remote have lost some of the data packets )
 * ----------------------------------------------------------------------------
 */
{
	time_t rpostime;
	
	if (txfp == NULL) {
		debug(DBG_JANUS, "xrcvrpos: error, rpos received but tx file is closed");
		TX_ERROR(xwaitpkt);
	};

	if (txstartsize == -1L) {
		debug(DBG_JANUS, "xrcvrpos: error, rpos received but transfer hasn't started");
		TX_ERROR(xwaitpkt);
	};
	
	if (rxbsize != 8) {
		debug(DBG_JANUS, "xrcvrpos: invalid RPOS packet received, size %ld", rxbsize);
		TX_ERROR(xwaitpkt);
	};
	
	rpostime = LSB_TO_MACHINE_32(*((UINT32 *) rxbuf + sizeof(UINT32)));
	if (rpostime != lastrpostime) {
		/* If it is't the rpos packet with same time - reposition */
		lastrpostime = rpostime;
		fseek(txfp, LSB_TO_MACHINE_32(*((UINT32 *) rxbuf)), SEEK_SET);
	
                /* Stop waiting for ACK*/
                if (restartstate == xsendblk) RESETTIMER(ACK_TIMER);
 
		/* Decrease blk size */
		txblksize >>= 1; sentbytes = 0;
		blkstepbytes += JBLKSIZE / 2;
                if (blkstepbytes > JBLKSIZE * 2)
                        blkstepbytes = JBLKSIZE * 2;
		if (txblksize < 64) 
			txblksize = 64;
	};
	TX_PROCEED(xsendblk);
	
}

TX_STATE(xrcveofack)
/* ----------------------------------------------------------------------------
 * receive EOF packet. Remote wants us to move to next file
 * ----------------------------------------------------------------------------
 */
	if (txfp == NULL) {
		debug(DBG_JANUS, "xrceofack error, eofack received but tx file is closed");
		TX_ERROR(xwaitpkt);
	};

	if (txstartsize == -1L) {
		debug(DBG_JANUS, "xrcvrpos: error, rpos received but transfer hasn't started");
		TX_ERROR(xwaitpkt);
	};
	
#ifdef IFCS
	ipcloginf(5, "janus: sent %lu bytes, %lu seconds, %lu CPS",
		(unsigned long) txsize - txstartsize, (unsigned long) difftime(time(NULL),txtime),
		(unsigned long) ( ( txsize - txstartsize ) / difftime(time(NULL),txtime) ) );
	if(shmData) 
	{
		shmData->send.TrTime += difftime(time(NULL),txtime);
		shmData->send.Start = 0L;
		strcpy(shmData->send.FileName, "");
		shmData->send.Pos = 0L;
		shmData->send.FileSize = 0L;
		shmData->send.SkipSize = 0L;
		shmData->send.BlockSize = 0;
		shmData->send.TrFiles+=(txsize - txstartsize);
	}
#endif
	ackcount = 0;
	fclose(txfp); txfp = NULL;
	LOGFILE(tx, "sent");

	execute_disposition(tmpfl); 
	tmpfl = tmpfl->next; 
	
	TX_PROCEED(xsendfname);
	
TX_STATE(xrcvfreq)
/* ----------------------------------------------------------------------------
 * receive FREQ packet. Remote asks us to send some file rather then ending 
 * batch 
 * ----------------------------------------------------------------------------
 */
{
	file_list *freq_files, *p;
	
	remotecaps = *((UCHAR *) rxbuf + 1 + strlen(rxbuf));
	debug(DBG_JANUS, "xrcvfreq: remote caps = %x",  remotecaps);

	/* Warning: respond_bark modifies it's argument */
	freq_files = respond_bark(rxbuf); 

	if (freq_files) {
		if (tosend) {
			/* This is better to be this way. We can receive FREQ packet
			 * at any time
			 */
			for (p = tosend; p -> next != NULL; p = p -> next); 
			p -> next = freq_files;
		} else {
			tosend = freq_files;
		};
		if (!tmpfl) tmpfl = freq_files;
		TX_PROCEED(xsendfname);
	} else {
		TX_PROCEED(xsendfreqnak);
	};
}	

TX_STATE(xsendfreqnak)
/* ----------------------------------------------------------------------------
 * send FREQNAK packet. Files, that remote wanted were not found 
 * ----------------------------------------------------------------------------
 */
	sendpkt(0, JPKT_FREQNAK, txbuf, 0);
	/* Set timeout for ack */
	SETTIMER(ACK_TIMER, ACK_TIMEOUT); restartstate = xsendfreqnak;
	TX_PROCEED(xwaitpkt);
	
TX_STATE(xrcvfrqnakack)
/* ----------------------------------------------------------------------------
 * receive FREQNAK ACK packet - continue processing files 
 * ----------------------------------------------------------------------------
 */
	if (txfp != NULL) {
		debug(DBG_JANUS, "xrcvfrnakack error, frnakack received in the middle of file transfer");
		TX_ERROR(xwaitpkt);
	};
	
	ackcount = 0;
	TX_PROCEED(xsendfname);

	
/* ============================================================================
 *
 * Receiver part
 *
 * ============================================================================
 */
	
RX_STATE(rrcvfname)
/* ----------------------------------------------------------------------------
 * receive FNAME packet. Remote want to transfer some stuff
 * ----------------------------------------------------------------------------
 */
	if (rxfp != NULL) {
		closefile(0);
		rxfp = NULL;
	};
	
	if (rxbsize >= 2) {
		/* Find remote caps if any */
		char *p;
		p = strchr(rxbuf, '\0') + 1;
		p = strchr(p,     '\0') + 1;
		remotecaps = rxbsize > p - rxbuf ? *((UCHAR *) p) : 0;
		
	} else {
		remotecaps = 0;
	};
	debug(DBG_JANUS, "rrcvblk: remote caps = %x",  remotecaps);	

	if (*rxbuf != '\0') { 
		/* Normal filename */
		size_t size;
		unsigned long remsize, remtime;
		unsigned int  remattr;
		
		size = strlen(rxbuf);
		if (sscanf(rxbuf + size + 1, "%lu %lo %o", &remsize, 
			&remtime, &remattr) == 3) {
				
			rxsize = remsize;
			rxfp = openfile(rxbuf, (time_t) remtime, (off_t) 
					remsize, &rxstartsize, NULL);
		        loginf("janus receive: \"%s\" %ld bytes dated %s mode %o",
			          rxbuf, remsize, date(remtime), remattr);
#ifdef IFCS
			ipcloginf(6, "janus: receive \"%s\" %ld bytes dated %s",
			          rxbuf, remsize, date(remtime));
			if(shmData) 
	  		{
	    			if(snprintf(shmData->recv.FileName, PATH_MAX, rxbuf)<0)
				{
					shmData->recv.FileName[PATH_MAX]='>';
					shmData->recv.FileName[PATH_MAX]='\0';
				}
	    			shmData->recv.FileSize = remsize;
	    			shmData->recv.Start = time (NULL);
	    			if(!shmData->recv.TrStart) shmData->recv.TrStart = time (NULL);
	  		}
#endif		
			if (!rxfp) {
				logerr("janus: file open error");
#ifdef IFCS
				ipcloginf (7, "janus: file open error");
#endif
				BSM_ERROR;
			};
			if (rxstartsize > 0) {
			        loginf("janus resyncing at offset: %lu", rxstartsize);
#ifdef IFCS
			 	ipcloginf(5, "janus: resyncing at offset: %lu", rxstartsize);
#endif		
			};

		} else {
			logerr("janus: bad fname packet received");
#ifdef IFCS
		 	ipcloginf(7,"janus: bad fname packet received");
#endif		
			rxstartsize = -1L;
		};
	} else {
		/* NULL filename.Remote has nothing to send */
		if (reqfiles) {
			/* We havn't got capabilities yet */
			if (remotecaps & JCAP_FREQ) {
				char *file;
			        file = strsep(&reqp, "\n\r");
				if (file) 
				if (*file) {
					loginf("janus: requesting file %s", file);
#ifdef IFCS
				 	ipcloginf(5,"janus: requesting file %s", file);
#endif		
					txbsize = sprintf(txbuf, "%.*s%c%c", JTXBUFSIZE-2,
						       	file, 0, JCAP_OUR);
					sendpkt(0, JPKT_FREQ, txbuf, txbsize += 2);
					RX_PROCEED(rwaitpkt);
				};
			};
		};
		debug(DBG_JANUS, "NULL block size: %lu, data : %s", rxbsize, printable(rxbuf, rxbsize));
		rxstartsize = 0L;
	};

	*((UINT32 *) txbuf) = MACHINE_TO_LSB_32(rxstartsize);
	*((UCHAR *) txbuf + 4) = JCAP_OUR /* | JCAP_CLEAR8 */;
	sendpkt(0, JPKT_FNACK, txbuf, 5);
	ackcount = 0;
	
	(void)time(&rxtime);

	if (*rxbuf != '\0' || tmpfl) {
		/* If it's a file or we have something to send - don't end batch*/
		/* T-Mail sometimes sends some files  after sending NULL pkt */
		RX_PROCEED(rwaitpkt);
	} else {
		RX_SUCCESS(rwaitpkt);/***/
	};
	
RX_STATE(rrcvblk)
/* ----------------------------------------------------------------------------
 * receive data packet. 
 * ----------------------------------------------------------------------------
 */
	if (!rxfp) {
		debug(DBG_JANUS, "rrcvblk: blk received but rx file is closed");
		sendpkt(0, JPKT_EOFACK, NULL, 0);
		ackcount++; /* This waiting has to stop sometime too */
		RX_PROCEED(rwaitpkt);
	};
	
	if (LSB_TO_MACHINE_32(*((UINT32 *)rxbuf)) != ftell(rxfp)) {
		/* If it's got too far - send repos pkt */
		if (EXPIRED(RPOS_TIMER)) {
			*((UINT32 *) txbuf) = MACHINE_TO_LSB_32(ftell(rxfp));
			*((UINT32 *) rxbuf + sizeof(UINT32)) = MACHINE_TO_LSB_32(time(NULL));
			sendpkt(0, JPKT_RPOS, txbuf, 4);
			SETTIMER(RPOS_TIMER, RPOS_TIMEOUT);
		};
	    	RX_PROCEED(rwaitpkt);
	};
		
	RESETTIMER(RPOS_TIMER);
		
	fwrite(rxbuf + sizeof(UINT32), 1, rxbsize - sizeof(UINT32), rxfp);
#ifdef IFCS
	if(shmData) 
	{
		if (ftell(rxfp) != rxsize) shmData->recv.Pos = ftell(rxfp);
		shmData->recv.BlockSize = rxbsize - sizeof(UINT32);
	}
#endif
	if (ftell(rxfp) == rxsize) {
	/* file received ok */
#ifdef IFCS
		ipcloginf(5, "janus: received %lu bytes, %lu seconds, %lu CPS",
			(unsigned long) rxsize - rxstartsize, (unsigned long) difftime(time(NULL),rxtime),
			(unsigned long) ( ( rxsize - rxstartsize ) / difftime(time(NULL),rxtime) ) );
		if(shmData) 
		{
			shmData->recv.TrTime += difftime(time(NULL), rxtime);
			shmData->recv.Start = 0L;
			strcpy(shmData->recv.FileName, "");
			shmData->recv.Pos = 0L;
			shmData->recv.FileSize = 0L;
			shmData->recv.SkipSize = 0L;
			shmData->recv.BlockSize = 0;
			shmData->recv.TrFiles+=rxsize - rxstartsize;
		}
#endif		
		closefile(1); rxfp = NULL;
		LOGFILE(rx, "received");
		sendpkt(0, JPKT_EOFACK, NULL, 0);
	};
	
	RX_PROCEED(rwaitpkt);
	
RX_STATE(rrcvfreqnak)
/* ----------------------------------------------------------------------------
 * receive freqnak packet. 
 * ----------------------------------------------------------------------------
 */
	loginf("janus: freq was declined");
#ifdef IFCS
 	ipcloginf(7,"janus: freq was declined");
#endif		
	sendpkt(0, JPKT_FRNAKACK, txbuf, 0);
	RX_PROCEED(rwaitpkt);
	
BSM_END
	/* Cleanup */
	
	if (rxfp != NULL) {
		/* Last file wasn't received completely - close it */
		rxsize = ftell(rxfp);
		closefile(0);
		LOGFILE(rx, "received");

	};
	if (txfp != NULL) {
		txsize = ftell(txfp);
		fclose(txfp); txfp = NULL;
		LOGFILE(tx, "dropped after");
	};
	
	if (SUCCESS && !(tx_success == -1 || rx_success == -1)) {
		if (end_janus_session(pkttype == JPKT_HALT))
			rc = 0;
		else
			rc = 2;
	} else
		rc = 2;

	loginf("janus: %s transfer", rc == 0 ? "end" : "abort");
#ifdef IFCS
 	ipcloginf(1,"janus: %s transfer", rc == 0 ? "end" : "abort");
	if(shmData)
	{
		shmData->send.TrStart=0;
		shmData->send.TrSize=0L;
		shmData->send.TrSkip=0L;
		shmData->send.TrFiles=0L;
		shmData->recv.TrStart=0;
		shmData->recv.TrSize=0L;
		shmData->recv.TrSkip=0L;
		shmData->recv.TrFiles=0L;
	}
#endif		
	tidy_filelist(tosend, rc == 0);
	
	if (reqfiles) { 
		/* We've done all reqs*/
		if (!strsep(&reqp, "\r\n")) {
			unlink(reqname(eff_remote->addr));
		};
		free(reqfiles);
	};
	
	free(pkt_buffer);
	free(txbuf);
	free(rxbuf);

	return rc;	
}

/*
 * End janus session  
 */
int end_janus_session(int haltreceived)
{
	int i, done, pkt;
	size_t rxbsize;
	char *rxbuf;
	

SM_STATES("end_janus_session")
	waitpkt, rejectpkt, haltpkt, haltackpkt, timeoutrcv
SM_EDECL
	rxbuf = xmalloc(JRXBUFSIZE + sizeof(unsigned long) + sizeof(int));
	done  = haltreceived ? 1 : 0; rxbsize = 0L;
	SETTIMER(DEAD_TIMER, DEAD_TIMEOUT);

	if (done) sendpkt(0, JPKT_HALTACK, NULL, 0); 
SM_START(rejectpkt)

SM_STATE(waitpkt)
	/* Remote not responding */
	if (EXPIRED(DEAD_TIMER)) { SM_ERROR; };

	if (done == 3) { 
	/* We've received HALT and HALTACK pkt - OK */
		for (i = 0; i < 10; i++) 
			sendpkt(0, JPKT_HALTACK, NULL, 0);
		/* Flush it up */
		FLUSHIN();
		FLUSHOUT();
		
		SM_SUCCESS; 
	};
	
	pkt = rcvpkt(&rxbsize, rxbuf, RECV_TIMEOUT);
	
	if (!SUCCESS) { SM_ERROR }; 

	if (pkt != JPKT_NONE && pkt != JPKT_BAD) {
		/* If we receive good packet it gives us a chance to wait more */
		RESETTIMER(DEAD_TIMER);
	};
	
	/* Handle interesting packets (Don't try to use switch !!! )*/
	if (pkt == JPKT_HALT) {
		SM_PROCEED(haltpkt);
	} else if (pkt == JPKT_HALTACK) {
		SM_PROCEED(haltackpkt);
	};
	

	SM_PROCEED(rejectpkt);
	
SM_STATE(rejectpkt)
	sendpkt(0, JPKT_HALT, NULL, 0);
	SM_PROCEED(waitpkt);
	
SM_STATE(haltpkt)
	sendpkt(0, JPKT_HALTACK, NULL, 0);
	done |= 1;
	SM_PROCEED(waitpkt);
	
SM_STATE(haltackpkt)
	done |= 2;
	SM_PROCEED(waitpkt);
SM_END
	free(rxbuf);
	return sm_success;
}



/*
 * Open and lock file.   
 */
extern FILE *flockopen(const char *fname, const char *mode)
{
	struct flock fl;
	FILE *fp;

	fl.l_whence=0;
	fl.l_start=0L;
	fl.l_len=0L;
	
	if ((fp = fopen(fname, mode)) == NULL) {
		int serrno = errno;
		logerr("cannot open file %s, skipping", S(fname));
#ifdef IFCS
	  	ipcloginf (7, "janus: cannot open file %s, skipping", S(fname));
#endif
		errno = serrno;
		return NULL;
	};

	/* Determine lock type */
	for (fl.l_type=F_RDLCK; *mode; mode++) 
		if (*mode == '+' || *mode == 'w' || *mode == 'a') {
			fl.l_type=F_WRLCK;
			break;
		};

	if (fcntl(fileno(fp), F_SETLK, &fl)) {
		loginf("cannot lock file %s, skipping", S(fname));
#ifdef IFCS
	  	ipcloginf (7, "janus: cannot lock file %s, skipping", S(fname));
#endif
		fclose(fp);
		return NULL;
	};
	return fp;
			
}

/*
 * Read FREQS list into memory if any
 * All sorts of checks ;)
 */

char *read_freqs(fa_list *remote)
{
	FILE	*rqfile;
	size_t	fsize;
	char	*names = NULL, *req;

	req = reqname(remote->addr);
	if (req) {
		rqfile = fopen(req, "rt");
		if (rqfile) {
			fseek(rqfile, 0, SEEK_END);
			fsize = ftell(rqfile);
			fseek(rqfile, 0, SEEK_SET);
			if (!ferror(rqfile)) {
				names = xmalloc(fsize + 1);
				if (names) {
					if (fread(names, fsize, 1, rqfile)) {
						names[fsize] = '\0';
					} else {
						free(names);
						names = NULL;
					};
				};
			};	
			fclose(rqfile);
		};
	};
	return names;
}

/* send janus packet 
 * for speed we accumulate it in pkt_buffer and then send in one 
 * call to PUT. to eluminate size checks allocate pkt_buffer as twice as
 * max janus packet could be
 */
/* FIXME: if for some reason write is blocked because of some problems 
 * program is locked too. Need to figure out a solution
 */
static int sendpkt(caps, type, buf, len)
int caps;
int type;
char *buf;
int len;
{
	register int clear8, crc32;
	char *p, *q;
	unsigned long crc;
	
	clear8 = 0; /*caps & JCAP_CLEAR8;*/
	crc32  = 0; /*caps & JCAP_CRC32;*/
	
	debug(DBG_JANUS,"sendpkt%s '%s' (%d)", crc32 ? "32" : "16",
		printablec(type),len);
	
	p = pkt_buffer; crc = crc32 ? 0xFFFFFFFFL : 0;
	/* Send header */
	*p++ = DLE;
	*p++ = (crc32 ? JDLM_32START : JDLM_START) ^ 0x40;
	/* Fill our buffer */
	if (crc32) {
		for (q = buf; len > 0; q++, len--) {
			crc = upd_crc32(*q, crc);
			escape_byte(clear8, *q, &p);
		};
		crc = upd_crc32(type, crc);
	} else {
		for (q = buf; len > 0; q++, len--) {
			crc = upd_crc16xmodem(*q, crc);
			escape_byte(clear8, *q, &p);
		};
		crc = upd_crc16xmodem(type, crc);
	};
	escape_byte(clear8, type, &p);

	*p++ = DLE;
	*p++ = JDLM_END ^ 0x40;
	if (crc32) {
		escape_byte(clear8, crc >> 24, &p);  
		escape_byte(clear8, crc >> 16, &p);	
	};
	escape_byte(clear8, crc >> 8, &p);  
	escape_byte(clear8, crc, &p);
	return PUT(pkt_buffer, p - pkt_buffer) == STAT_SUCCESS;

}

/* Put escaped byte into buffer p
 * XON, XOFF, DLE and "CR after @" are escaped
 */
static void escape_byte(int clear8, char b, char **p)
{
	b &= 0xFF;
	switch (b) {
		case CR:
				if (*(*p - 1) != '@') break;
		case DC1:
		case DC3:
				if (clear8) break;
		case DLE:
				*(*p)++ = DLE;
				b ^= 0x40;
				break;
		default:
				break;
	};
	*(*p)++ = b;
}


/* receive janus packet, return type or -1 on error 
   WARNING: We use tail of buffer to store CRC and it's type ! 
   and it has to be big enough to hold type + one char 
   of possible garbage also !!! */

/*TODO: would be good to get all available data at once instead of 
  multiple GET's. but how ? */
static int rcvpkt(int *size, char *buf, int timeout)
{
	int i, c = 0;
	unsigned long rcvcrc, *crc = (unsigned long *) (buf + JRXBUFSIZE);
	int *crctype = (int *) (buf + JRXBUFSIZE + sizeof(unsigned long));
	char *ptr;
	/* Try to find start of the buffer if any */
	if (!*size) while ((c = get_unescaped_byte(timeout)) >= 0) ; 
	/* Read data till end marker is found or size limit is exceeded */
	if (*size || c == JPKT_START || c == JPKT_32START) {
		if (c /* or !*size*/) {
			*crctype = c;
		        *crc = c == JPKT_START ? 0L : 0xFFFFFFFFL;
		};
		/* Duplicated for speed */
		if (*crctype == JPKT_START) {
			for (ptr = buf + *size; *size < JRXBUFSIZE; (*size)++) {
				if ((c = get_unescaped_byte(timeout)) < 0) break; 
				*ptr++ = c; *crc = upd_crc16xmodem(c, (*crc)) & 0xFFFFL;
				
			};
		} else {
			for (ptr = buf + *size; *size < JRXBUFSIZE; (*size)++) {
				if ((c = get_unescaped_byte(timeout)) < 0) break; 
				*ptr++ = c; *crc = upd_crc32(c, (*crc));
			};	
		};
		
	} else if (c == JPKT_END) {
		c = 0; /* Don't fool crc calculation below */		
	};
	
	switch (c) {
		case JPKT_END:
		/* it is right packet (we found end marker - fetch crc and compare */
			
			for (rcvcrc = 0, i = 0; i < (*crctype == JPKT_START ? 2 : 4); i++) {
				if ((c = get_unescaped_byte(RECV_TIMEOUT)) >= 0) {
					/* TODO: do it for MSB */
					rcvcrc <<= 8;
					rcvcrc |= c & 0xFF; 
				} else {
				       	debug(DBG_JANUS, "crc%s timeout %c", *crctype == JPKT_START ? "16" : "32",*(buf + (*size) - 1)); 
					*size = 0;
					return JPKT_BAD;
				};
			};
			if (*crc == rcvcrc) {
	/*			if (time(NULL) % 3) {
					*size = 0;
					return JPKT_BAD;
				};*/
				debug(DBG_JANUS, "rcvpkt: pkt %c, size: %lu", *(buf + (*size) - 1), *size - 1);
				return *(buf + --(*size));
			} else {
				debug(DBG_JANUS, "rcvpkt: received packet type %x. local crc %x remote crc %x", c, *crc, rcvcrc);
				*size = 0;
				return  JPKT_BAD;
			};
		case TIMEOUT: case EMPTY:
			return JPKT_NONE;
		case HANGUP: case ERROR:
			*size = 0;
			return JPKT_HALT;
		default:
			debug(DBG_JANUS, "rcvpkt: filled up buffer and have got %x at end", c);
			*size = 0;
			return JPKT_BAD;
	};
}

static int get_unescaped_byte(int timeout)
{
	int c = GETCHAR(timeout);
	if (c == DLE) {
		if ((c = GETCHAR(timeout)) >= 0) c ^= 0x40;
		/* Check for JDLM* . they are normal chars, so they can easilly be
		 * distinguished from usual packet data
		 */
		switch (c) {
			case JDLM_START:	return JPKT_START;
			case JDLM_END:		return JPKT_END;
			case JDLM_32START:	return JPKT_32START;
			default:			return c;
		};
		
	};
	return c;
}
